import re
import typing

import edq.util.json

import lms.model.base
import lms.util.string

T = typing.TypeVar('T')

class BaseQuery(edq.util.json.DictConverter):
    """
    Queries are ways that users can attempt to refer to some object with uncertainty.
    This allows users to refer to objects by name, for example, instead of by id.

    Queries are made up of 2-3 components:
     - an identifier
     - a name
     - an email (optional)

    Email support is decided by child classes.
    By default, ids are assumed to be only digits.

    A query can be represented in text the following ways:
     - LMS ID (`id`)
     - Email (`email`)
     - Full Name (`name`)
     - f"{email} ({id})"
     - f"{name} ({id})"
    """

    _include_email: bool = True
    """ Control if this class instance supports the email field. """

    def __init__(self,
            id: typing.Union[str, int, None] = None,
            name: typing.Union[str, None] = None,
            email: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        if (id is not None):
            id = str(id)

        self.id: typing.Union[str, None] = id
        """ The LMS's identifier for this query. """

        self.name: typing.Union[str, None] = name
        """ The display name of this query. """

        self.email: typing.Union[str, None] = email
        """ The email address of this query. """

        if ((self.id is None) and (self.name is None) and (self.email is None)):
            raise ValueError("Query is empty, it must have at least one piece of information (id, name, email).")

    def match(self, target: typing.Union[typing.Any, 'BaseQuery', None]) -> bool:
        """
        Check if this query matches the given target.
        A missing field in the query means that field will not be checked.
        A missing field in the target is seen as empty and mill by checked against.
        """

        if (target is None):
            return False

        field_names = ['id', 'name']
        if (self._include_email):
            field_names.append('email')

        for field_name in field_names:
            self_value = getattr(self, field_name, None)
            target_value = getattr(target, field_name, None)

            if (self_value is None):
                continue

            if (self_value != target_value):
                return False

        return True

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        data = {
            'id': self.id,
            'name': self.name,
        }

        if (self._include_email):
            data['email'] = self.email

        return data

    def _get_comparison_payload(self, include_id: bool) -> typing.Tuple:
        """ Get values for comparison. """

        payload = []

        if (include_id):
            payload.append(self.id)

        payload.append(self.name)

        if (self._include_email):
            payload.append(self.email)

        return tuple(payload)

    def __eq__(self, other: object) -> bool:
        if (not isinstance(other, BaseQuery)):
            return False

        # Check the ID specially.
        comparison = lms.util.string.compare_maybe_ints(self.id, other.id)
        if (comparison != 0):
            return False

        return self._get_comparison_payload(False) == other._get_comparison_payload(False)

    def __lt__(self, other: object) -> bool:
        if (not isinstance(other, BaseQuery)):
            return False

        # Check the ID specially.
        comparison = lms.util.string.compare_maybe_ints(self.id, other.id)
        if (comparison != 0):
            return (comparison < 0)

        return self._get_comparison_payload(False) < other._get_comparison_payload(False)

    def __hash__(self) -> int:
        return hash(self._get_comparison_payload(True))

    def __str__(self) -> str:
        text = self.email
        if ((not self._include_email) or (text is None)):
            text = self.name

        if (self.id is not None):
            if (text is not None):
                text = f"{text} ({self.id})"
            else:
                text = self.id

        if (text is None):
            return '<unknown>'

        return text

    def _to_text(self) -> str:
        """ Represent this query as a string. """

        return str(self)

class ResolvedBaseQuery(BaseQuery):
    """
    A BaseQuery that has been resolved (verified) from a real instance.
    """

    def __init__(self,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        if (self.id is None):
            raise ValueError("A resolved query cannot be created without an ID.")

    def get_id(self) -> str:
        """ Get the ID (which must exists) for this query. """

        if (self.id is None):
            raise ValueError("A resolved query cannot be created without an ID.")

        return self.id

def parse_int_query(query_type: typing.Type[T], text: typing.Union[str, None],
        check_email: bool = True,
        ) -> typing.Union[T, None]:
    """
    Parse a query with the assumption that LMS ids are ints.

    Accepts queries are in the following forms:
        - LMS ID (`id`)
        - Email (`email`)
        - Name (`name`)
        - f"{email} ({id})"
        - f"{name} ({id})"
    """

    if (text is None):
        return None

    # Clean whitespace.
    text = re.sub(r'\s+', ' ', str(text)).strip()
    if (len(text) == 0):
        return None

    id = None
    email = None
    name = None

    match = re.search(r'^(\S.*)\((\d+)\)$', text)
    if (match is not None):
        # Query has both text and id.
        name = match.group(1).strip()
        id = match.group(2)
    elif (re.search(r'^\d+$', text) is not None):
        # Query must be an ID.
        id = text
    else:
        name = text

    # Check if the name is actually an email address.
    if (check_email and (name is not None) and ('@' in name)):
        email = name
        name = None

    data = {
        'id': id,
        'name': name,
        'email': email,
    }

    return query_type(**data)



import copy

import lms.backend.testing
import lms.model.courses
import lms.model.groups
import lms.model.groupsets
import lms.model.users
import lms.model.testdata.groups

def test_courses_groupsets_memberships_resolve_and_subtract_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of resolving and subtracting group set memberships. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # No Op
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'memberships': [
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-1'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-3'),
                    ),
                ]
            },
            {
                lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(): 0,
            },
            None,
        ),

        # Sub Only
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'memberships': [
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-1'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-1'),
                    ),
                ]
            },
            {
                lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(): 1,
            },
            None,
        ),

        # All
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'memberships': [
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-1'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-1'),
                    ),
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-2'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-1'),
                    ),
                ]
            },
            {
                lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(): 1,
                lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-2'].to_query(): 0,
            },
            None,
        ),
    ]

    def _assertion_func(expected, actual):
        # Convert the non-JSON keys.

        clean_expected = {}
        for (key, value) in expected.items():
            clean_expected[str(key)] = value

        clean_actual = {}
        for (key, value) in actual.items():
            clean_actual[str(key)] = value

        test.assertDictEqual(clean_expected, clean_actual)

    test.base_request_test(test.backend.courses_groupsets_memberships_resolve_and_subtract, test_cases,
            assertion_func = _assertion_func)



# pylint: disable=abstract-method

import typing

import edq.util.net

import lms.backend.blackboard.model
import lms.model.backend
import lms.model.constants
import lms.model.courses
import lms.model.users
import lms.util.net
import lms.util.parse

class BlackboardBackend(lms.model.backend.APIBackend):
    """ An API backend for the Blackboard Learn LMS. """

    def __init__(self,
            server: str,
            auth_user: typing.Union[str, None] = None,
            auth_password: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(server, lms.model.constants.BACKEND_TYPE_BLACKBOARD, **kwargs)

        if (auth_user is None):
            raise ValueError("Blackboard backends require a username.")

        if (auth_password is None):
            raise ValueError("Blackboard backends require a password.")

        self._username = auth_user
        """ The username to authenticate with. """

        self._password = auth_password
        """ The password to authenticate with. """

        self._session_headers: typing.Union[typing.Dict[str, typing.Any], None] = None
        """ The headers (e.g., cookies) for our logged in Blackboard session. """

    def _login(self) -> None:
        """ Try to login to the Blackboard server. """

        # Check if we are already logged in.
        if (self._session_headers is not None):
            return

        response, _ = edq.util.net.make_get(self.server)
        cookies, router_params = self._parse_bb_cookies(response.headers.get('set-cookie', None))

        new_cookies = {
            'BbRouter': cookies['bbrouter']
        }

        text_cookies = '; '.join(['='.join(items) for items in new_cookies.items()])

        headers = {
            'cookie': text_cookies,
        }
        data = {
            'user_id': self._username,
            'password': self._password,
            'blackboard.platform.security.NonceUtil.nonce.ajax': router_params['xsrf'],
        }

        response, _ = edq.util.net.make_post(self.server + '/webapps/login/',
                headers = headers, data = data,
                # Don't store the nonce in exchanges.
                params_to_skip = ['blackboard.platform.security.NonceUtil.nonce.ajax'],
                allow_redirects = False)

        cookies, router_params = self._parse_bb_cookies(response.headers.get('set-cookie', None))
        if ('sessionId' not in router_params):
            raise ValueError(f"Could not log into Blackboard server ({self.server}) with user '{self._username}'. Is username/password correct?")

        self._session_headers = {
            'cookie': response.headers.get('set-cookie', None),
            'x-blackboard-xsrf': router_params['xsrf'],
            # Insert a header to identify the user.
            'edq-lms-blackboard-user': self._username,
        }

    def _parse_bb_cookies(self, text_cookies: typing.Union[str, None]) -> typing.Tuple[typing.Dict[str, typing.Any], typing.Dict[str, str]]:
        """
        Parse out the Blackboard cookies and return two dicts:
         - All cookies (with lower case keys).
         - The router params/cookies (specific to Blackboard).
        """

        cookies: typing.Dict[str, typing.Any] = {}
        router_params: typing.Dict[str, str] = {}

        # If we are testing, return fake cookies.
        if (self.is_testing()):
            cookies['bbrouter'] = 'xsrf:abc,sessionId:9999999'
            router_params['xsrf'] = 'abc'
            router_params['sessionId'] = '9999999'

            return cookies, router_params

        cookies = lms.util.net.parse_cookies(text_cookies)

        router_text = cookies.get('bbrouter', None)
        if (isinstance(router_text, str)):
            for text in router_text.split(','):
                key, value = text.split(':', maxsplit = 1)
                router_params[key] = value

        return cookies, router_params

    def courses_list(self,
            **kwargs: typing.Any) -> typing.List[lms.model.courses.Course]:
        self._login()

        url = self.server + '/learn/api/public/v3/courses'
        data = {
            'availability.available': 'Yes',
        }
        response, _ = edq.util.net.make_get(url, headers = self._session_headers, data = data)

        courses = []
        for raw_course in response.json().get('results', []):
            courses.append(lms.backend.blackboard.model.course(raw_course))

        courses.sort()

        return courses

    def courses_users_list(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.users.CourseUser]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')

        self._login()

        url = self.server + '/learn/api/public/v1/courses/{course_id}/users'
        url = url.format(course_id = lms.backend.blackboard.model.format_id(parsed_course_id))

        data = {
            'expand': 'user',
        }

        response, _ = edq.util.net.make_get(url, headers = self._session_headers, data = data)

        users = []
        for raw_user in response.json().get('results', []):
            users.append(lms.backend.blackboard.model.course_user(raw_user))

        users.sort()

        return users



import typing

import lms.backend.canvas.common
import lms.model.constants

BASE_ENDPOINT = "/api/v1/group_categories/{groupset_id}"

def request(backend: typing.Any,
        course_id: int,
        groupset_id: int,
        ) -> bool:
    """ Delete a group set. """

    url = backend.server + BASE_ENDPOINT.format(groupset_id = groupset_id)
    headers = backend.get_standard_headers()

    headers[lms.model.constants.HEADER_KEY_WRITE] = 'true'

    raw_object = lms.backend.canvas.common.make_delete_request(url, headers = headers)
    if (raw_object is None):
        identifiers = {
            'course_id': course_id,
            'groupset_id': groupset_id,
        }
        backend.not_found('delete group set', identifiers)

        return False

    return True



"""
Subtract users from a group set.
"""

import argparse
import sys

import lms.backend.instance
import lms.cli.common
import lms.cli.courses.groupsets.memberships.common
import lms.cli.parser
import lms.model.base

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    config = args._config

    backend = lms.backend.instance.get_backend(**config)

    course_query = lms.cli.common.check_required_course(backend, config)
    if (course_query is None):
        return 1

    groupset_query = lms.cli.common.check_required_groupset(backend, config)
    if (groupset_query is None):
        return 2

    memberships = lms.cli.courses.groupsets.memberships.common.load_group_memberships(backend, args.path, args.skip_rows)

    counts = backend.courses_groupsets_memberships_resolve_and_subtract(course_query, groupset_query, memberships)

    for (group_query, count) in counts.items():
        print(f"Subtracted {count} users from group {group_query}.")

    return 0

def main() -> int:
    """ Get a parser, parse the args, and call run. """
    return run_cli(_get_parser().parse_args())

def _get_parser() -> argparse.ArgumentParser:
    """ Get the parser. """

    parser = lms.cli.parser.get_parser(__doc__.strip(),
            include_output_format = True,
            include_course = True,
            include_groupset = True,
            include_group = True,
            include_skip_rows = True,
    )

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to a TSV file where each row has 2 columns: group query and user query.')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import typing

import lms.backend.canvas.common
import lms.backend.canvas.model
import lms.model.scores

BASE_ENDPOINT = "/api/v1/courses/{course_id}/assignments/{assignment_id}/submissions"

def request(backend: typing.Any,
        course_id: int,
        assignment_id: int,
        ) -> typing.List[lms.model.scores.AssignmentScore]:
    """ List assignment scores. """

    url = backend.server + BASE_ENDPOINT.format(course_id = course_id, assignment_id = assignment_id)
    headers = backend.get_standard_headers()

    raw_objects = lms.backend.canvas.common.make_get_request_list(url, headers = headers)
    if (raw_objects is None):
        identifiers = {
            'course_id': course_id,
            'assignment_id': assignment_id,
        }
        backend.not_found('list assignment scores', identifiers)

        return []

    scores = []
    for raw_object in raw_objects:
        # Check if this is an actual submission and not just a placeholder.
        if (raw_object.get('workflow_state', None) == 'unsubmitted'):
            continue

        scores.append(lms.backend.canvas.model.assignment_score(raw_object))

    return scores



import lms.backend.testing
import lms.model.testdata.groupsets

def test_courses_groupsets_fetch_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of fetching course groupsets. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_id': '130000000',
                'groupset_id': '131010100',
            },
            lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'],
            None,
        ),

        # Miss
        (
            {
                'course_id': '110000000',
                'groupset_id': '999',
            },
            None,
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_groupsets_fetch, test_cases)



import os
import typing

import edq.util.dirent

import lms.backend.canvas.common
import lms.model.constants

BASE_ENDPOINT = "/api/v1/group_categories/{groupset_id}/import"

def request(backend: typing.Any,
        course_id: int,
        groupset_id: int,
        group_id: int,
        user_ids: typing.List[int],
        ) -> int:
    """ Add users to a group. """

    url = backend.server + BASE_ENDPOINT.format(groupset_id = groupset_id)
    headers = backend.get_standard_headers()

    headers[lms.model.constants.HEADER_KEY_WRITE] = 'true'

    # Build a CSV.
    rows = ["canvas_user_id,canvas_group_id"]
    for user_id in sorted(user_ids):
        rows.append(f"{user_id},{group_id}")

    # Write CSV.
    temp_dir = edq.util.dirent.get_temp_dir('edq-lms-canvas-')
    temp_path = os.path.join(temp_dir, 'data.csv')
    edq.util.dirent.write_file_bytes(temp_path, "\n".join(rows).encode(edq.util.dirent.DEFAULT_ENCODING))

    files = {
        'attachment': open(temp_path, 'rb'),  # pylint: disable=consider-using-with
    }

    raw_object = lms.backend.canvas.common.make_post_request(url, headers = headers, files = files)
    if (raw_object is None):
        identifiers = {
            'course_id': course_id,
            'groupset_id': groupset_id,
            'group_id': group_id,
            'user_ids': user_ids,
        }
        backend.not_found('add group memberships', identifiers)

        return 0

    return len(user_ids)



import logging
import typing

import edq.util.parse

import lms.model.assignments
import lms.model.constants
import lms.model.courses
import lms.model.groups
import lms.model.groupsets
import lms.model.query
import lms.model.scores
import lms.model.users

class APIBackend():
    """
    API backends provide a unified interface to an LMS.

    Note that instead of using an abstract class,
    methods will raise a NotImplementedError by default.
    This will allow child backends to fill in as much functionality as they can,
    while still leaving gaps where they are incomplete or impossible.
    """

    _testing_override: typing.Union[bool, None] = None
    """ A top-level override to control testing status. """

    def __init__(self,
            server: str,
            backend_type: str,
            testing: typing.Union[bool, str] = False,
            **kwargs: typing.Any) -> None:
        self.server: str = server
        """ The server this backend will connect to. """

        self.backend_type: str = backend_type
        """
        The type for this backend.
        Should be set by the child class.
        """

        parsed_testing = edq.util.parse.boolean(testing)
        if (APIBackend._testing_override is not None):
            parsed_testing = APIBackend._testing_override

        self.testing: bool = parsed_testing
        """ True if the backend is being used for a test. """

    # Core Methods

    def is_testing(self) -> bool:
        """ Check if this backend is in testing mode. """

        return self.testing

    def get_standard_headers(self) -> typing.Dict[str, str]:
        """
        Get standard headers for this backend.
        Children should take care to set the write header when performing a write operation.
        """

        return {
            lms.model.constants.HEADER_KEY_BACKEND: self.backend_type,
            lms.model.constants.HEADER_KEY_WRITE: 'false',
        }

    def not_found(self, operation: str, identifiers: typing.Dict[str, typing.Any]) -> None:
        """
        Called when the backend was unable to find some object.
        This will only be called when a requested object is not found,
        e.g., a user requested by ID is not found.
        This is not called when a list naturally returns zero results,
        or when a query does not match any items.
        """

        logging.warning("Object not found during operation: '%s'. Identifiers: %s.", operation, identifiers)

    # API Methods

    def courses_get(self,
            course_queries: typing.Collection[lms.model.courses.CourseQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.courses.Course]:
        """
        Get the specified courses associated with the given course.
        """

        if (len(course_queries) == 0):
            return []

        courses = self.courses_list(**kwargs)

        matches = []
        for course in sorted(courses):
            for query in course_queries:
                if (query.match(course)):
                    matches.append(course)
                    break

        return sorted(matches)

    def courses_fetch(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.courses.Course, None]:
        """
        Fetch a single course associated with the context user.
        Return None if no matching course is found.

        By default, this will just do a list and choose the relevant record.
        Specific backends may override this if there are performance concerns.
        """

        courses = self.courses_list(**kwargs)
        for course in courses:
            if (course.id == course_id):
                return course

        return None

    def courses_list(self,
            **kwargs: typing.Any) -> typing.List[lms.model.courses.Course]:
        """
        List the courses associated with the context user.
        """

        raise NotImplementedError('courses_list')

    def courses_assignments_get(self,
            course_query: lms.model.courses.CourseQuery,
            assignment_queries: typing.Collection[lms.model.assignments.AssignmentQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.assignments.Assignment]:
        """
        Get the specified assignments associated with the given course.
        """

        if (len(assignment_queries) == 0):
            return []

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)

        assignments = sorted(self.courses_assignments_list(resolved_course_query.get_id(), **kwargs))
        assignment_queries = sorted(assignment_queries)

        matches = []
        for assignment in assignments:
            for query in assignment_queries:
                if (query.match(assignment)):
                    matches.append(assignment)
                    break

        return matches

    def courses_assignments_fetch(self,
            course_id: str,
            assignment_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.assignments.Assignment, None]:
        """
        Fetch a single assignment associated with the given course.
        Return None if no matching assignment is found.

        By default, this will just do a list and choose the relevant record.
        Specific backends may override this if there are performance concerns.
        """

        assignments = self.courses_assignments_list(course_id, **kwargs)
        for assignment in sorted(assignments):
            if (assignment.id == assignment_id):
                return assignment

        return None

    def courses_assignments_list(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.assignments.Assignment]:
        """
        List the assignments associated with the given course.
        """

        raise NotImplementedError('courses_assignments_list')

    def courses_assignments_resolve_and_list(self,
            course_query: lms.model.courses.CourseQuery,
            **kwargs: typing.Any) -> typing.List[lms.model.assignments.Assignment]:
        """
        List the assignments associated with the given course.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        return sorted(self.courses_assignments_list(resolved_course_query.get_id(), **kwargs))

    def courses_assignments_scores_get(self,
            course_query: lms.model.courses.CourseQuery,
            assignment_query: lms.model.assignments.AssignmentQuery,
            user_queries: typing.Collection[lms.model.users.UserQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.scores.AssignmentScore]:
        """
        Get the scores associated with the given assignment query and user queries.
        """

        if (len(user_queries) == 0):
            return []

        scores = self.courses_assignments_scores_resolve_and_list(course_query, assignment_query, **kwargs)

        matches = []
        for score in scores:
            for user_query in user_queries:
                if (user_query.match(score.user_query)):
                    matches.append(score)

        return sorted(matches)

    def courses_assignments_scores_fetch(self,
            course_id: str,
            assignment_id: str,
            user_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.scores.AssignmentScore, None]:
        """
        Fetch the score associated with the given assignment and user.

        By default, this will just do a list and choose the relevant record.
        Specific backends may override this if there are performance concerns.
        """

        scores = self.courses_assignments_scores_list(course_id, assignment_id, **kwargs)
        for score in scores:
            if ((score.user_query is not None) and (score.user_query.id == user_id)):
                return score

        return None

    def courses_assignments_scores_list(self,
            course_id: str,
            assignment_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.scores.AssignmentScore]:
        """
        List the scores associated with the given assignment.
        """

        raise NotImplementedError('courses_assignments_scores_list')

    def courses_assignments_scores_resolve_and_list(self,
            course_query: lms.model.courses.CourseQuery,
            assignment_query: lms.model.assignments.AssignmentQuery,
            **kwargs: typing.Any) -> typing.List[lms.model.scores.AssignmentScore]:
        """
        List the scores associated with the given assignment query.
        In addition to resolving the assignment query,
        users will also be resolved into their full version
        (instead of the reduced version usually returned with scores).
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)

        # Resolve the assignment query.
        matched_assignments = self.courses_assignments_get(resolved_course_query, [assignment_query], **kwargs)
        if (len(matched_assignments) == 0):
            return []

        target_assignment = matched_assignments[0]

        # List the scores.
        scores = self.courses_assignments_scores_list(resolved_course_query.get_id(), target_assignment.id, **kwargs)
        if (len(scores) == 0):
            return []

        # Resolve the scores' queries.

        users = self.courses_users_list(resolved_course_query.get_id(), **kwargs)
        users_map = {user.id: user for user in users}

        for score in scores:
            score.assignment_query = target_assignment.to_query()

            if ((score.user_query is not None) and (score.user_query.id in users_map)):
                score.user_query = users_map[score.user_query.id].to_query()

        return sorted(scores)

    def courses_assignments_scores_resolve_and_upload(self,
            course_query: lms.model.courses.CourseQuery,
            assignment_query: lms.model.assignments.AssignmentQuery,
            scores: typing.Dict[lms.model.users.UserQuery, lms.model.scores.ScoreFragment],
            **kwargs: typing.Any) -> int:
        """
        Resolve queries and upload assignment scores (indexed by user query).
        A None score (ScoreFragment.score) indicates that the score should be cleared.
        Return the number of scores sent to the LMS.
        """

        if (len(scores) == 0):
            return 0

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_assignment_query = self.resolve_assignment_query(resolved_course_query.get_id(), assignment_query, **kwargs)

        resolved_users = self.resolve_user_queries(resolved_course_query.get_id(), list(scores.keys()), warn_on_miss = True)
        resolved_scores: typing.Dict[str, lms.model.scores.ScoreFragment] = {}

        for (user, score) in scores.items():
            for resolved_user in resolved_users:
                if (user.match(resolved_user)):
                    resolved_scores[resolved_user.get_id()] = score
                    continue

        if (len(resolved_scores) == 0):
            return 0

        return self.courses_assignments_scores_upload(
                resolved_course_query.get_id(),
                resolved_assignment_query.get_id(),
                resolved_scores,
                **kwargs)

    def courses_assignments_scores_upload(self,
            course_id: str,
            assignment_id: str,
            scores: typing.Dict[str, lms.model.scores.ScoreFragment],
            **kwargs: typing.Any) -> int:
        """
        Upload assignment scores (indexed by user id).
        A None score (ScoreFragment.score) indicates that the score should be cleared.
        Return the number of scores sent to the LMS.
        """

        raise NotImplementedError('courses_assignments_scores_upload')

    def courses_gradebook_get(self,
            course_query: lms.model.courses.CourseQuery,
            assignment_queries: typing.Collection[lms.model.assignments.AssignmentQuery],
            user_queries: typing.Collection[lms.model.users.UserQuery],
            **kwargs: typing.Any) -> lms.model.scores.Gradebook:
        """
        Get a gradebook with the specified users and assignments.
        Specifying no users/assignments is the same as requesting all of them.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)

        resolved_assignment_queries = self.resolve_assignment_queries(resolved_course_query.get_id(), assignment_queries, empty_all = True, **kwargs)
        assignment_ids = [query.get_id() for query in resolved_assignment_queries]

        resolved_user_queries = self.resolve_user_queries(resolved_course_query.get_id(), user_queries,
                empty_all = True, only_students = True, **kwargs)
        user_ids = [query.get_id() for query in resolved_user_queries]

        gradebook = self.courses_gradebook_fetch(resolved_course_query.get_id(), assignment_ids, user_ids, **kwargs)

        # Resolve the gradebook's queries (so it can show names/emails instead of just IDs).
        gradebook.update_queries(resolved_assignment_queries, resolved_user_queries)

        return gradebook

    def courses_gradebook_fetch(self,
            course_id: str,
            assignment_ids: typing.Collection[str],
            user_ids: typing.Collection[str],
            **kwargs: typing.Any) -> lms.model.scores.Gradebook:
        """
        Get a gradebook with the specified users and assignments.
        If either the assignments or users is empty, an empty gradebook will be returned.
        """

        raise NotImplementedError('courses_gradebook_fetch')

    def courses_gradebook_list(self,
            course_id: str,
            **kwargs: typing.Any) -> lms.model.scores.Gradebook:
        """
        List the full gradebook associated with this course.
        """

        return self.courses_gradebook_get(lms.model.courses.CourseQuery(id = course_id), [], [], **kwargs)

    def courses_gradebook_resolve_and_list(self,
            course_query: lms.model.courses.CourseQuery,
            **kwargs: typing.Any) -> lms.model.scores.Gradebook:
        """
        List the full gradebook associated with this course.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        return self.courses_gradebook_list(resolved_course_query.get_id(), **kwargs)

    def courses_gradebook_resolve_and_upload(self,
            course_query: lms.model.courses.CourseQuery,
            gradebook: lms.model.scores.Gradebook,
            **kwargs: typing.Any) -> int:
        """
        Resolve queries and upload a gradebook.
        Missing scores in the gradebook are skipped,
        a None score (ScoreFragment.score) indicates that the score should be cleared.
        Return the number of scores sent to the LMS.
        """

        if (len(gradebook) == 0):
            return 0

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)

        assignments = self.courses_assignments_list(resolved_course_query.get_id(), **kwargs)
        users = self.courses_users_list(resolved_course_query.get_id(), **kwargs)

        resolved_assignment_queries = [assignment.to_query() for assignment in assignments]
        resolved_user_queries = [user.to_query() for user in users]

        gradebook.update_queries(resolved_assignment_queries, resolved_user_queries)

        return self.courses_gradebook_upload(
                resolved_course_query.get_id(),
                gradebook,
                **kwargs)

    def courses_gradebook_upload(self,
            course_id: str,
            gradebook: lms.model.scores.Gradebook,
            **kwargs: typing.Any) -> int:
        """
        Upload a gradebook.
        All queries in the gradebook must be resolved (or at least have an ID).
        Missing scores in the gradebook are skipped,
        a None score (ScoreFragment.score) indicates that the score should be cleared.
        Return the number of scores sent to the LMS.
        """

        assignment_scores = gradebook.get_scores_by_assignment()

        count = 0
        for (assignment, user_scores) in assignment_scores.items():
            if (assignment.id is None):
                raise ValueError(f"Assignment query for gradebook upload ({assignment}) does not have an ID.")

            upload_scores = {}
            for (user, score) in user_scores.items():
                if (user.id is None):
                    raise ValueError(f"User query for gradebook upload ({user}) does not have an ID.")

                upload_scores[user.id] = score.to_fragment()

            count += self.courses_assignments_scores_upload(course_id, assignment.id, upload_scores, **kwargs)

        return count

    def courses_groupsets_create(self,
            course_id: str,
            name: str,
            **kwargs: typing.Any) -> lms.model.groupsets.GroupSet:
        """
        Create a group set.
        """

        raise NotImplementedError('courses_groupsets_create')

    def courses_groupsets_resolve_and_create(self,
            course_query: lms.model.courses.CourseQuery,
            name: str,
            **kwargs: typing.Any) -> lms.model.groupsets.GroupSet:
        """
        Resolve references and create a group set.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        return self.courses_groupsets_create(resolved_course_query.get_id(), name, **kwargs)

    def courses_groupsets_delete(self,
            course_id: str,
            groupset_id: str,
            **kwargs: typing.Any) -> bool:
        """
        Delete a group set.
        """

        raise NotImplementedError('courses_groupsets_delete')

    def courses_groupsets_resolve_and_delete(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            **kwargs: typing.Any) -> bool:
        """
        Resolve references and create a group set.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)
        return self.courses_groupsets_delete(resolved_course_query.get_id(), resolved_groupset_query.get_id(), **kwargs)

    def courses_groupsets_get(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_queries: typing.Collection[lms.model.groupsets.GroupSetQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.GroupSet]:
        """
        Get the specified group sets associated with the given course.
        """

        if (len(groupset_queries) == 0):
            return []

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        groupset_queries = sorted(groupset_queries)
        groupsets = sorted(self.courses_groupsets_list(resolved_course_query.get_id(), **kwargs))

        matches = []
        for groupset in groupsets:
            for query in groupset_queries:
                if (query.match(groupset)):
                    matches.append(groupset)
                    break

        return matches

    def courses_groupsets_fetch(self,
            course_id: str,
            groupset_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.groupsets.GroupSet, None]:
        """
        Fetch a single group set associated with the given course.
        Return None if no matching group set is found.

        By default, this will just do a list and choose the relevant record.
        Specific backends may override this if there are performance concerns.
        """

        groupsets = self.courses_groupsets_list(course_id, **kwargs)
        for groupset in groupsets:
            if (groupset.id == groupset_id):
                return groupset

        return None

    def courses_groupsets_list(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.GroupSet]:
        """
        List the group sets associated with the given course.
        """

        raise NotImplementedError('courses_groupsets_list')

    def courses_groupsets_resolve_and_list(self,
            course_query: lms.model.courses.CourseQuery,
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.GroupSet]:
        """
        List the group sets associated with the given course.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        return sorted(self.courses_groupsets_list(resolved_course_query.get_id(), **kwargs))

    def courses_groupsets_memberships_resolve_and_add(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            memberships: typing.Collection[lms.model.groups.GroupMembership],
            **kwargs: typing.Any) -> typing.Tuple[
                    typing.List[lms.model.groups.Group],
                    typing.Dict[lms.model.groups.ResolvedGroupQuery, int]
            ]:
        """
        Resolve queries and add the specified users to the specified groups.
        This may create groups.

        Return:
         - Created Groups
         - Group Addition Counts
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)

        found_group_memberships, missing_group_memberships, _ = self._resolve_group_memberships(
                resolved_course_query.get_id(), resolved_groupset_query.get_id(), memberships, **kwargs)

        # Create missing groups.
        created_groups = []
        for name in sorted(missing_group_memberships.keys()):
            group = self.courses_groups_create(resolved_course_query.get_id(), resolved_groupset_query.get_id(), name, **kwargs)
            created_groups.append(group)

            # Merge in new group with existing structure.
            query = group.to_query()
            if (query not in found_group_memberships):
                found_group_memberships[query] = []

            found_group_memberships[query] += missing_group_memberships[name]

        # Add memberships.
        counts = {}
        for resolved_group_query in sorted(found_group_memberships.keys()):
            resolved_user_queries = found_group_memberships[resolved_group_query]

            count = self.courses_groups_memberships_resolve_and_add(
                    resolved_course_query, resolved_groupset_query, resolved_group_query,
                    resolved_user_queries,
                    **kwargs)

            counts[resolved_group_query] = count

        return (created_groups, counts)

    def courses_groupsets_memberships_resolve_and_set(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            memberships: typing.Collection[lms.model.groups.GroupMembership],
            **kwargs: typing.Any) -> typing.Tuple[
                    typing.List[lms.model.groups.Group],
                    typing.List[lms.model.groups.ResolvedGroupQuery],
                    typing.Dict[lms.model.groups.ResolvedGroupQuery, int],
                    typing.Dict[lms.model.groups.ResolvedGroupQuery, int],
            ]:
        """
        Resolve queries and set the specified group memberships.
        This may create and delete groups.

        Return:
         - Created Groups
         - Deleted Groups
         - Group Addition Counts
         - Group Subtraction Counts
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)

        found_group_memberships, missing_group_memberships, unused_groups = self._resolve_group_memberships(
                resolved_course_query.get_id(), resolved_groupset_query.get_id(), memberships, **kwargs)

        # Delete unused groups.
        deleted_groups = []
        for group_query in sorted(unused_groups):
            result = self.courses_groups_delete(resolved_course_query.get_id(), resolved_groupset_query.get_id(), group_query.get_id(), **kwargs)
            if (result):
                deleted_groups.append(group_query)

        # Create missing groups.
        created_groups = []
        for name in sorted(missing_group_memberships.keys()):
            group = self.courses_groups_create(resolved_course_query.get_id(), resolved_groupset_query.get_id(), name, **kwargs)
            created_groups.append(group)

            # Merge in new group with existing structure.
            query = group.to_query()
            if (query not in found_group_memberships):
                found_group_memberships[query] = []

            found_group_memberships[query] += missing_group_memberships[name]

        # Set memberships.
        add_counts = {}
        sub_counts = {}
        for resolved_group_query in sorted(found_group_memberships.keys()):
            resolved_user_queries = found_group_memberships[resolved_group_query]

            (add_count, sub_count, deleted) = self.courses_groups_memberships_resolve_and_set(
                    resolved_course_query, resolved_groupset_query, resolved_group_query,
                    resolved_user_queries,
                    delete_empty = True,
                    **kwargs)

            if (deleted):
                deleted_groups.append(resolved_group_query)

            add_counts[resolved_group_query] = add_count
            sub_counts[resolved_group_query] = sub_count

        return (created_groups, deleted_groups, add_counts, sub_counts)

    def courses_groupsets_memberships_resolve_and_subtract(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            memberships: typing.Collection[lms.model.groups.GroupMembership],
            **kwargs: typing.Any) -> typing.Dict[lms.model.groups.ResolvedGroupQuery, int]:
        """
        Resolve queries and subtract the specified users to the specified groups.
        This will not delete any groups.

        Return:
         - Group Subtraction Counts
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)

        found_group_memberships, missing_group_memberships, _ = self._resolve_group_memberships(
                resolved_course_query.get_id(), resolved_groupset_query.get_id(), memberships, **kwargs)

        # Warn about missing groups.
        for name in sorted(missing_group_memberships.keys()):
            logging.warning("Group does not exist: '%s'.", name)

        # Subtract memberships.
        counts = {}
        for resolved_group_query in sorted(found_group_memberships.keys()):
            resolved_user_queries = found_group_memberships[resolved_group_query]

            (count, _) = self.courses_groups_memberships_resolve_and_subtract(
                    resolved_course_query, resolved_groupset_query, resolved_group_query,
                    resolved_user_queries,
                    delete_empty = False,
                    **kwargs)

            counts[resolved_group_query] = count

        return counts

    def courses_groupsets_memberships_list(self,
            course_id: str,
            groupset_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.GroupSetMembership]:
        """
        List the membership of the group sets associated with the given course.
        """

        raise NotImplementedError('courses_groupsets_memberships_list')

    def courses_groupsets_memberships_resolve_and_list(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.GroupSetMembership]:
        """
        List the membership of the group sets associated with the given course.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)

        memberships = self.courses_groupsets_memberships_list(resolved_course_query.get_id(), resolved_groupset_query.get_id(), **kwargs)

        # Resolve memberships.

        users = self.courses_users_list(resolved_course_query.get_id(), **kwargs)
        groups = self.courses_groups_list(resolved_course_query.get_id(), resolved_groupset_query.get_id(), **kwargs)

        users_map = {user.id: user.to_query() for user in users}
        groups_map = {group.id: group.to_query() for group in groups}

        for membership in memberships:
            membership.update_queries(resolved_groupset_query, users = users_map, groups = groups_map)

        return sorted(memberships)

    def courses_groups_create(self,
            course_id: str,
            groupset_id: str,
            name: str,
            **kwargs: typing.Any) -> lms.model.groups.Group:
        """
        Create a group.
        """

        raise NotImplementedError('courses_groups_create')

    def courses_groups_resolve_and_create(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            name: str,
            **kwargs: typing.Any) -> lms.model.groups.Group:
        """
        Resolve references and create a group.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)
        return self.courses_groups_create(resolved_course_query.get_id(), resolved_groupset_query.get_id(), name, **kwargs)

    def courses_groups_delete(self,
            course_id: str,
            groupset_id: str,
            group_id: str,
            **kwargs: typing.Any) -> bool:
        """
        Delete a group.
        """

        raise NotImplementedError('courses_groups_delete')

    def courses_groups_resolve_and_delete(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            group_query: lms.model.groups.GroupQuery,
            **kwargs: typing.Any) -> bool:
        """
        Resolve references and create a group.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)
        resolved_group_query = self.resolve_group_query(resolved_course_query.get_id(), resolved_groupset_query.get_id(), group_query, **kwargs)
        return self.courses_groups_delete(resolved_course_query.get_id(), resolved_groupset_query.get_id(), resolved_group_query.get_id(), **kwargs)

    def courses_groups_get(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            group_queries: typing.Collection[lms.model.groups.GroupQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.groups.Group]:
        """
        Get the specified groups associated with the given course.
        """

        if (len(group_queries) == 0):
            return []

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)
        groups = self.courses_groups_list(resolved_course_query.get_id(), resolved_groupset_query.get_id(), **kwargs)

        group_queries = sorted(group_queries)
        groups = sorted(groups)

        matches = []
        for group in groups:
            for query in group_queries:
                if (query.match(group)):
                    matches.append(group)
                    break

        return matches

    def courses_groups_fetch(self,
            course_id: str,
            groupset_id: str,
            group_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.groups.Group, None]:
        """
        Fetch a single group associated with the given course.
        Return None if no matching group is found.

        By default, this will just do a list and choose the relevant record.
        Specific backends may override this if there are performance concerns.
        """

        groups = self.courses_groups_list(course_id, groupset_id, **kwargs)
        for group in groups:
            if (group.id == group_id):
                return group

        return None

    def courses_groups_list(self,
            course_id: str,
            groupset_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.groups.Group]:
        """
        List the groups associated with the given course.
        """

        raise NotImplementedError('courses_groups_list')

    def courses_groups_resolve_and_list(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            **kwargs: typing.Any) -> typing.List[lms.model.groups.Group]:
        """
        List the groups associated with the given course.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)
        return self.courses_groups_list(resolved_course_query.get_id(), resolved_groupset_query.get_id(), **kwargs)

    def courses_groups_memberships_add(self,
            course_id: str,
            groupset_id: str,
            group_id: str,
            user_ids: typing.Collection[str],
            **kwargs: typing.Any) -> int:
        """
        Add the specified users to the group.
        """

        raise NotImplementedError('courses_groups_memberships_add')

    def courses_groups_memberships_resolve_and_add(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            group_query: lms.model.groups.GroupQuery,
            user_queries: typing.Collection[lms.model.users.UserQuery],
            **kwargs: typing.Any) -> int:
        """
        Resolve queries and add the specified users to the group.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)
        resolved_group_query = self.resolve_group_query(resolved_course_query.get_id(), resolved_groupset_query.get_id(), group_query, **kwargs)
        resolved_user_queries = self.resolve_user_queries(resolved_course_query.get_id(), user_queries, warn_on_miss = True, **kwargs)

        # Get users already in this group.
        group_memberships = self.courses_groups_memberships_list(
                resolved_course_query.get_id(),
                resolved_groupset_query.get_id(),
                resolved_group_query.get_id(),
                **kwargs)

        group_user_ids = {membership.user.id for membership in group_memberships if membership.user.id is not None}

        # Filter out users already in the group.
        user_ids = []
        for query in sorted(resolved_user_queries):
            if (query.get_id() in group_user_ids):
                logging.warning("User '%s' already in group '%s'.", query, resolved_group_query)
                continue

            user_ids.append(query.get_id())

        if (len(user_ids) == 0):
            return 0

        return self.courses_groups_memberships_add(
                resolved_course_query.get_id(),
                resolved_groupset_query.get_id(),
                resolved_group_query.get_id(),
                user_ids,
                **kwargs)

    def courses_groups_memberships_list(self,
            course_id: str,
            groupset_id: str,
            group_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.GroupSetMembership]:
        """
        List the membership of the group associated with the given group set.
        """

        raise NotImplementedError('courses_groups_memberships_list')

    def courses_groups_memberships_resolve_and_list(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            group_query: lms.model.groups.GroupQuery,
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.GroupSetMembership]:
        """
        List the membership of the group associated with the given group set.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)

        groups = self.courses_groups_get(resolved_course_query, resolved_groupset_query, [group_query], **kwargs)
        if (len(groups) == 0):
            raise ValueError(f"Unable to find group: '{group_query}'.")

        group = groups[0]

        memberships = self.courses_groups_memberships_list(
                resolved_course_query.get_id(),
                resolved_groupset_query.get_id(),
                group.id,
                **kwargs)

        # Resolve memberships.

        users = self.courses_users_list(resolved_course_query.get_id(), **kwargs)
        users_map = {user.id: user.to_query() for user in users}

        groups_map = {group.id: group.to_query()}

        for membership in memberships:
            membership.update_queries(resolved_groupset_query, users = users_map, groups = groups_map)

        return sorted(memberships)

    def courses_groups_memberships_resolve_and_set(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            group_query: lms.model.groups.GroupQuery,
            user_queries: typing.Collection[lms.model.users.UserQuery],
            delete_empty: bool = False,
            **kwargs: typing.Any) -> typing.Tuple[int, int, bool]:
        """
        Resolve queries and set the specified users for the group.
        This method can both add and subtract users from the group.

        Returns:
         - The count of users added.
         - The count of users subtracted.
         - If this group was deleted.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)
        resolved_group_query = self.resolve_group_query(resolved_course_query.get_id(), resolved_groupset_query.get_id(), group_query, **kwargs)
        resolved_user_queries = self.resolve_user_queries(resolved_course_query.get_id(), user_queries, warn_on_miss = True, **kwargs)

        # Get users already in this group.
        group_memberships = self.courses_groups_memberships_list(
                resolved_course_query.get_id(),
                resolved_groupset_query.get_id(),
                resolved_group_query.get_id(),
                **kwargs)

        group_user_queries = {membership.user for membership in group_memberships if membership.user is not None}
        group_user_ids = {membership.user.id for membership in group_memberships if membership.user.id is not None}
        query_user_ids = {resolved_user_query.get_id() for resolved_user_query in resolved_user_queries}

        # Collect users that need to be added.
        add_user_ids = []
        for query_user_id in query_user_ids:
            if (query_user_id not in group_user_ids):
                add_user_ids.append(query_user_id)

        # Collect users that need to be subtracted.
        sub_user_queries = []
        for group_user_query in group_user_queries:
            if (group_user_query not in resolved_user_queries):
                sub_user_queries.append(group_user_query)

        # Update the group.

        add_count = 0
        if (len(add_user_ids) > 0):
            add_count = self.courses_groups_memberships_add(
                    resolved_course_query.get_id(),
                    resolved_groupset_query.get_id(),
                    resolved_group_query.get_id(),
                    add_user_ids,
                    **kwargs)

        sub_count = 0
        deleted = False
        if (len(sub_user_queries) > 0):
            sub_count, deleted = self.courses_groups_memberships_resolve_and_subtract(
                    resolved_course_query,
                    resolved_groupset_query,
                    resolved_group_query,
                    sub_user_queries,
                    delete_empty = delete_empty,
                    **kwargs)

        return add_count, sub_count, deleted

    def courses_groups_memberships_subtract(self,
            course_id: str,
            groupset_id: str,
            group_id: str,
            user_ids: typing.Collection[str],
            **kwargs: typing.Any) -> int:
        """
        Subtract the specified users from the group.
        """

        raise NotImplementedError('courses_groups_memberships_subtract')

    def courses_groups_memberships_resolve_and_subtract(self,
            course_query: lms.model.courses.CourseQuery,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            group_query: lms.model.groups.GroupQuery,
            user_queries: typing.Collection[lms.model.users.UserQuery],
            delete_empty: bool = False,
            **kwargs: typing.Any) -> typing.Tuple[int, bool]:
        """
        Resolve queries and subtract the specified users from the group.
        Return:
            - The number of users deleted.
            - If this group was deleted.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        resolved_groupset_query = self.resolve_groupset_query(resolved_course_query.get_id(), groupset_query, **kwargs)
        resolved_group_query = self.resolve_group_query(resolved_course_query.get_id(), resolved_groupset_query.get_id(), group_query, **kwargs)
        resolved_user_queries = self.resolve_user_queries(resolved_course_query.get_id(), user_queries, warn_on_miss = True, **kwargs)

        # Get users already in this group.
        group_memberships = self.courses_groups_memberships_list(
                resolved_course_query.get_id(),
                resolved_groupset_query.get_id(),
                resolved_group_query.get_id(),
                **kwargs)

        group_user_ids = {membership.user.id for membership in group_memberships if membership.user.id is not None}

        # Filter out users not in the group.
        user_ids = []
        for query in resolved_user_queries:
            if (query.get_id() not in group_user_ids):
                logging.warning("User '%s' is not in group '%s'.", query, resolved_group_query)
                continue

            user_ids.append(query.get_id())

        if (delete_empty and len(group_memberships) == 0):
            deleted = self.courses_groups_delete(
                resolved_course_query.get_id(),
                resolved_groupset_query.get_id(),
                resolved_group_query.get_id(),
                **kwargs)
            return 0, deleted

        if (len(user_ids) == 0):
            return 0, False

        count = self.courses_groups_memberships_subtract(
                resolved_course_query.get_id(),
                resolved_groupset_query.get_id(),
                resolved_group_query.get_id(),
                user_ids,
                **kwargs)

        deleted = False
        if (delete_empty and (count == len(group_memberships))):
            deleted = self.courses_groups_delete(
                resolved_course_query.get_id(),
                resolved_groupset_query.get_id(),
                resolved_group_query.get_id(),
                **kwargs)

        return count, deleted

    def courses_syllabus_fetch(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.Union[str, None]:
        """
        Get the syllabus for a course, or None if no syllabus exists.
        """

        raise NotImplementedError('courses_syllabus_fetch')

    def courses_syllabus_get(self,
            course_query: lms.model.courses.CourseQuery,
            **kwargs: typing.Any) -> typing.Union[str, None]:
        """
        Get the syllabus for a course query, or None if no syllabus exists.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)

        return self.courses_syllabus_fetch(resolved_course_query.get_id(), **kwargs)

    def courses_users_get(self,
            course_query: lms.model.courses.CourseQuery,
            user_queries: typing.Collection[lms.model.users.UserQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.users.CourseUser]:
        """
        Get the specified users associated with the given course.
        """

        if (len(user_queries) == 0):
            return []

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        users = self.courses_users_list(resolved_course_query.get_id(), **kwargs)

        user_queries = sorted(user_queries)
        users = sorted(users)

        matches = []
        for user in users:
            for query in user_queries:
                if (query.match(user)):
                    matches.append(user)
                    break

        return matches

    def courses_users_fetch(self,
            course_id: str,
            user_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.users.CourseUser, None]:
        """
        Fetch a single user associated with the given course.
        Return None if no matching user is found.

        By default, this will just do a list and choose the relevant record.
        Specific backends may override this if there are performance concerns.
        """

        users = self.courses_users_list(course_id, **kwargs)
        for user in sorted(users):
            if (user.id == user_id):
                return user

        return None

    def courses_users_list(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.users.CourseUser]:
        """
        List the users associated with the given course.
        """

        raise NotImplementedError('courses_users_list')

    def courses_users_resolve_and_list(self,
            course_query: lms.model.courses.CourseQuery,
            **kwargs: typing.Any) -> typing.List[lms.model.users.CourseUser]:
        """
        List the users associated with the given course.
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)
        return list(sorted(self.courses_users_list(resolved_course_query.get_id(), **kwargs)))

    def courses_users_scores_get(self,
            course_query: lms.model.courses.CourseQuery,
            user_query: lms.model.users.UserQuery,
            assignment_queries: typing.Collection[lms.model.assignments.AssignmentQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.scores.AssignmentScore]:
        """
        Get the scores associated with the given user query and assignment queries.
        """

        if (len(assignment_queries) == 0):
            return []

        scores = self.courses_users_scores_resolve_and_list(course_query, user_query, **kwargs)

        scores = sorted(scores)
        assignment_queries = sorted(assignment_queries)

        matches = []
        for score in scores:
            for assignment_query in assignment_queries:
                if (assignment_query.match(score.assignment_query)):
                    matches.append(score)

        return matches

    def courses_users_scores_fetch(self,
            course_id: str,
            user_id: str,
            assignment_id: str,
            **kwargs: typing.Any) -> typing.Union[lms.model.scores.AssignmentScore, None]:
        """
        Fetch the score associated with the given user and assignment.

        By default, this will just do a list and choose the relevant record.
        Specific backends may override this if there are performance concerns.
        """

        # The default implementation is the same as courses_assignments_scores_fetch().
        return self.courses_assignments_scores_fetch(course_id, assignment_id, user_id, **kwargs)

    def courses_users_scores_list(self,
            course_id: str,
            user_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.scores.AssignmentScore]:
        """
        List the scores associated with the given user.
        """

        raise NotImplementedError('courses_users_scores_list')

    def courses_users_scores_resolve_and_list(self,
            course_query: lms.model.courses.CourseQuery,
            user_query: lms.model.users.UserQuery,
            **kwargs: typing.Any) -> typing.List[lms.model.scores.AssignmentScore]:
        """
        List the scores associated with the given user query.
        In addition to resolving the user query,
        assignments will also be resolved into their full version
        (instead of the reduced version usually returned with scores).
        """

        resolved_course_query = self.resolve_course_query(course_query, **kwargs)

        # Resolve the user query.
        matched_users = self.courses_users_get(resolved_course_query, [user_query], **kwargs)
        if (len(matched_users) == 0):
            return []

        target_user = matched_users[0]

        # List the scores.
        scores = self.courses_users_scores_list(resolved_course_query.get_id(), target_user.id, **kwargs)
        if (len(scores) == 0):
            return []

        # Resolve the scores' queries.

        assignments = self.courses_assignments_list(resolved_course_query.get_id(), **kwargs)
        assignments_map = {assignment.id: assignment for assignment in assignments}

        for score in scores:
            score.user_query = target_user.to_query()

            if ((score.assignment_query is not None) and (score.assignment_query.id in assignments_map)):
                score.assignment_query = assignments_map[score.assignment_query.id].to_query()

        return sorted(scores)

    # Utility Methods

    def parse_assignment_query(self, text: typing.Union[str, None]) -> typing.Union[lms.model.assignments.AssignmentQuery, None]:
        """
        Attempt to parse an assignment query from a string.
        The there is no query, return a None.
        If the query is malformed, raise an exception.

        By default, this method assumes that LMS IDs are ints.
        Child backends may override this to implement their specific behavior.
        """

        return lms.model.query.parse_int_query(lms.model.assignments.AssignmentQuery, text, check_email = False)

    def parse_assignment_queries(self, texts: typing.Collection[typing.Union[str, None]]) -> typing.List[lms.model.assignments.AssignmentQuery]:
        """ Parse a list of assignment queries. """

        queries = []
        for text in texts:
            query = self.parse_assignment_query(text)
            if (query is not None):
                queries.append(query)

        return queries

    def parse_course_query(self, text: typing.Union[str, None]) -> typing.Union[lms.model.courses.CourseQuery, None]:
        """
        Attempt to parse a course query from a string.
        The there is no query, return a None.
        If the query is malformed, raise an exception.

        By default, this method assumes that LMS IDs are ints.
        Child backends may override this to implement their specific behavior.
        """

        return lms.model.query.parse_int_query(lms.model.courses.CourseQuery, text, check_email = False)

    def parse_course_queries(self, texts: typing.Collection[typing.Union[str, None]]) -> typing.List[lms.model.courses.CourseQuery]:
        """ Parse a list of course queries. """

        queries = []
        for text in texts:
            query = self.parse_course_query(text)
            if (query is not None):
                queries.append(query)

        return queries

    def parse_groupset_query(self, text: typing.Union[str, None]) -> typing.Union[lms.model.groupsets.GroupSetQuery, None]:
        """
        Attempt to parse a group set query from a string.
        The there is no query, return a None.
        If the query is malformed, raise an exception.

        By default, this method assumes that LMS IDs are ints.
        Child backends may override this to implement their specific behavior.
        """

        return lms.model.query.parse_int_query(lms.model.groupsets.GroupSetQuery, text, check_email = False)

    def parse_groupset_queries(self, texts: typing.Collection[typing.Union[str, None]]) -> typing.List[lms.model.groupsets.GroupSetQuery]:
        """ Parse a list of group set queries. """

        queries = []
        for text in texts:
            query = self.parse_groupset_query(text)
            if (query is not None):
                queries.append(query)

        return queries

    def parse_group_query(self, text: typing.Union[str, None]) -> typing.Union[lms.model.groups.GroupQuery, None]:
        """
        Attempt to parse a group query from a string.
        The there is no query, return a None.
        If the query is malformed, raise an exception.

        By default, this method assumes that LMS IDs are ints.
        Child backends may override this to implement their specific behavior.
        """

        return lms.model.query.parse_int_query(lms.model.groups.GroupQuery, text, check_email = False)

    def parse_group_queries(self, texts: typing.Collection[typing.Union[str, None]]) -> typing.List[lms.model.groups.GroupQuery]:
        """ Parse a list of group queries. """

        queries = []
        for text in texts:
            query = self.parse_group_query(text)
            if (query is not None):
                queries.append(query)

        return queries

    def parse_user_query(self, text: typing.Union[str, None]) -> typing.Union[lms.model.users.UserQuery, None]:
        """
        Attempt to parse a user query from a string.
        The there is no query, return a None.
        If the query is malformed, raise an exception.

        By default, this method assumes that LMS IDs are ints.
        Child backends may override this to implement their specific behavior.
        """

        return lms.model.query.parse_int_query(lms.model.users.UserQuery, text, check_email = True)

    def parse_user_queries(self, texts: typing.Collection[typing.Union[str, None]]) -> typing.List[lms.model.users.UserQuery]:
        """ Parse a list of user queries. """

        queries = []
        for text in texts:
            query = self.parse_user_query(text)
            if (query is not None):
                queries.append(query)

        return queries

    def resolve_assignment_query(self,
            course_id: str,
            assignment_query: lms.model.assignments.AssignmentQuery,
            **kwargs: typing.Any) -> lms.model.assignments.ResolvedAssignmentQuery:
        """ Resolve the assignment query or raise an exception. """

        # Shortcut already resolved queries.
        if (isinstance(assignment_query, lms.model.assignments.ResolvedAssignmentQuery)):
            return assignment_query

        results = self.resolve_assignment_queries(course_id, [assignment_query], **kwargs)
        if (len(results) == 0):
            raise ValueError(f"Could not resolve assignment query: '{assignment_query}'.")

        return results[0]

    def resolve_assignment_queries(self,
            course_id: str,
            queries: typing.Collection[lms.model.assignments.AssignmentQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.assignments.ResolvedAssignmentQuery]:
        """
        Resolve a list of assignment queries into a list of resolved assignment queries.
        See _resolve_queries().
        """

        results = self._resolve_queries(
            queries,
            'assignment',
            self.courses_assignments_list(course_id, **kwargs),
            lms.model.assignments.ResolvedAssignmentQuery,
            **kwargs)

        return typing.cast(typing.List[lms.model.assignments.ResolvedAssignmentQuery], results)

    def resolve_course_query(self,
            query: lms.model.courses.CourseQuery,
            **kwargs: typing.Any) -> lms.model.courses.ResolvedCourseQuery:
        """ Resolve the course query or raise an exception. """

        # Shortcut already resolved queries.
        if (isinstance(query, lms.model.courses.ResolvedCourseQuery)):
            return query

        results = self.resolve_course_queries([query], **kwargs)
        if (len(results) == 0):
            raise ValueError(f"Could not resolve course query: '{query}'.")

        return results[0]

    def resolve_course_queries(self,
            queries: typing.Collection[lms.model.courses.CourseQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.courses.ResolvedCourseQuery]:
        """
        Resolve a list of course queries into a list of resolved course queries.
        See _resolve_queries().
        """

        results = self._resolve_queries(
            queries,
            'course',
            self.courses_list(**kwargs),
            lms.model.courses.ResolvedCourseQuery,
            **kwargs)

        return typing.cast(typing.List[lms.model.courses.ResolvedCourseQuery], results)

    def resolve_group_queries(self,
            course_id: str,
            groupset_id: str,
            queries: typing.Collection[lms.model.groups.GroupQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.groups.ResolvedGroupQuery]:
        """
        Resolve a list of group queries into a list of resolved group queries.
        See _resolve_queries().
        """

        results = self._resolve_queries(
            queries,
            'group',
            self.courses_groups_list(course_id, groupset_id, **kwargs),
            lms.model.groups.ResolvedGroupQuery,
            **kwargs)

        return typing.cast(typing.List[lms.model.groups.ResolvedGroupQuery], results)

    def resolve_group_query(self,
            course_id: str,
            groupset_id: str,
            query: lms.model.groups.GroupQuery,
            **kwargs: typing.Any) -> lms.model.groups.ResolvedGroupQuery:
        """ Resolve the group query or raise an exception. """

        # Shortcut already resolved queries.
        if (isinstance(query, lms.model.groups.ResolvedGroupQuery)):
            return query

        results = self.resolve_group_queries(course_id, groupset_id, [query], **kwargs)
        if (len(results) == 0):
            raise ValueError(f"Could not resolve group query: '{query}'.")

        return results[0]

    def resolve_groupset_queries(self,
            course_id: str,
            queries: typing.Collection[lms.model.groupsets.GroupSetQuery],
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.ResolvedGroupSetQuery]:
        """
        Resolve a list of group set queries into a list of resolved group set queries.
        See _resolve_queries().
        """

        results = self._resolve_queries(
            queries,
            'group set',
            self.courses_groupsets_list(course_id, **kwargs),
            lms.model.groupsets.ResolvedGroupSetQuery,
            **kwargs)

        return typing.cast(typing.List[lms.model.groupsets.ResolvedGroupSetQuery], results)

    def resolve_groupset_query(self,
            course_id: str,
            groupset_query: lms.model.groupsets.GroupSetQuery,
            **kwargs: typing.Any) -> lms.model.groupsets.ResolvedGroupSetQuery:
        """ Resolve the group set query or raise an exception. """

        # Shortcut already resolved queries.
        if (isinstance(groupset_query, lms.model.groupsets.ResolvedGroupSetQuery)):
            return groupset_query

        results = self.resolve_groupset_queries(course_id, [groupset_query], **kwargs)
        if (len(results) == 0):
            raise ValueError(f"Could not resolve group set query: '{groupset_query}'.")

        return results[0]

    def resolve_user_queries(self,
            course_id: str,
            queries: typing.Collection[lms.model.users.UserQuery],
            only_students: bool = False,
            **kwargs: typing.Any) -> typing.List[lms.model.users.ResolvedUserQuery]:
        """
        Resolve a list of user queries into a list of resolved user queries.
        See _resolve_queries().
        """

        filter_func = None
        if (only_students):
            filter_func = lambda user: user.is_student()  # pylint: disable=unnecessary-lambda-assignment

        results = self._resolve_queries(
            queries,
            'user',
            self.courses_users_list(course_id, **kwargs),
            lms.model.users.ResolvedUserQuery,
            filter_func = filter_func,
            **kwargs)

        return typing.cast(typing.List[lms.model.users.ResolvedUserQuery], results)

    def _resolve_queries(self,
            queries: typing.Collection[lms.model.query.BaseQuery],
            label: str,
            items: typing.Collection,
            resolved_query_class: typing.Type,
            empty_all: bool = False,
            warn_on_miss: bool = False,
            filter_func: typing.Union[typing.Callable, None] = None,
            **kwargs: typing.Any) -> typing.List[lms.model.query.ResolvedBaseQuery]:
        """
        Resolve a list of queries.
        The returned list may be shorter than the list of queries (if input queries are not matched).
        The queries will be deduplicated and sorted.

        If |empty_all| is true and no queries are specified, then all items will be returned.

        If |filter_func| is passed, then that function will be called with each raw item,
        and ones that return true will be kept.
        """

        if (filter_func is not None):
            items = list(filter(filter_func, items))

        if (empty_all and (len(queries) == 0)):
            return list(sorted({resolved_query_class(item) for item in items}))

        matched_queries = []  # type: ignore[var-annotated]
        for query in queries:
            match = False
            for item in items:
                if (query.match(item)):
                    matched_query = resolved_query_class(item)

                    if (match):
                        raise ValueError(
                            f"Ambiguous {label} query ('{query}')"
                            f" matches multiple {label}s ['{matched_queries[-1]}', '{matched_query}'].")

                    matched_queries.append(matched_query)
                    match = True

            if ((not match) and warn_on_miss):
                logging.warning("Could not resolve %s query '%s'.", label, query)

        return list(sorted(set(matched_queries)))

    def _resolve_group_memberships(self,
            course_id: str,
            groupset_id: str,
            memberships: typing.Collection[lms.model.groups.GroupMembership],
            **kwargs: typing.Any) -> typing.Tuple[
                typing.Dict[lms.model.groups.ResolvedGroupQuery, typing.List[lms.model.users.ResolvedUserQuery]],
                typing.Dict[str, typing.List[lms.model.users.ResolvedUserQuery]],
                typing.List[lms.model.groups.ResolvedGroupQuery]]:
        """
        Resolve a list of group memberships.
        This method will resolved each query and split up the memberships by the appropriate group.
        If a group does not exist, the memberships will be split by apparent group name.

        Returns:
         - Memberships in Found Groups (keyed by resolved group query)
         - Memberships in Missing Groups (keyed by apparent group name)
         - Groups not involved in any of the returned memberships.

        The returned dicts will be the found groups (keyed by resolved query) and then the missing groups (keyed by apparent group name).
        """

        found_group_memberships: typing.Dict[lms.model.groups.ResolvedGroupQuery, typing.List[lms.model.users.ResolvedUserQuery]] = {}
        missing_group_memberships: typing.Dict[str, typing.List[lms.model.users.ResolvedUserQuery]] = {}

        users = self.courses_users_list(course_id, **kwargs)
        resolved_user_queries = [user.to_query() for user in sorted(users)]

        groups = self.courses_groups_list(course_id, groupset_id, **kwargs)
        resolved_group_queries = [group.to_query() for group in sorted(groups)]

        for (i, membership) in enumerate(memberships):
            # Resolve user.

            resolved_user_query = None
            for possible_user_query in resolved_user_queries:
                if (membership.user.match(possible_user_query)):
                    resolved_user_query = possible_user_query
                    break

            if (resolved_user_query is None):
                logging.warning("Could not resolve user '%s' for membership entry at index %d.", membership.user, i)
                continue

            # Resolve group.

            resolved_group_query = None
            for possible_group_query in resolved_group_queries:
                if (membership.group.match(possible_group_query)):
                    resolved_group_query = possible_group_query
                    break

            # Add to the correct collection.

            if (resolved_group_query is None):
                if ((membership.group.name is None) or (len(membership.group.name) == 0)):
                    logging.warning(("Membership entry at index %d has a group with no name."
                        + " Ensure that non-existent groups all have names."), i)
                    continue

                if (membership.group.name not in missing_group_memberships):
                    missing_group_memberships[membership.group.name] = []

                missing_group_memberships[membership.group.name].append(resolved_user_query)
            else:
                if (resolved_group_query not in found_group_memberships):
                    found_group_memberships[resolved_group_query] = []

                found_group_memberships[resolved_group_query].append(resolved_user_query)

        unused_groups = sorted(list(set(resolved_group_queries) - set(found_group_memberships.keys())))

        return (found_group_memberships, missing_group_memberships, unused_groups)



import typing

import edq.util.dirent

import lms.model.backend
import lms.model.groups

def load_group_memberships(
        backend: lms.model.backend.APIBackend,
        path: str,
        skip_rows: bool,
        ) -> typing.List[lms.model.groups.GroupMembership]:
    """ Read a group membership TSV file. """

    memberships: typing.List[lms.model.groups.GroupMembership] = []

    with open(path, 'r', encoding = edq.util.dirent.DEFAULT_ENCODING) as file:
        lineno = 0
        real_rows = 0
        for line in file:
            lineno += 1

            if (line.strip() == ''):
                continue

            real_rows += 1

            if (real_rows <= skip_rows):
                continue

            parts = [part.strip() for part in line.split("\t")]
            if (len(parts) != 2):
                raise ValueError(f"File '{path}' line {lineno} has the incorrect number of values. Expecting 2, found {len(parts)}.")

            group_query = backend.parse_group_query(parts[0])
            if (group_query is None):
                raise ValueError(f"File '{path}' line {lineno} has a group query that could not be parsed: '{parts[0]}'.")

            user_query = backend.parse_user_query(parts[1])
            if (user_query is None):
                raise ValueError(f"File '{path}' line {lineno} has a user query that could not be parsed: '{parts[1]}'.")

            memberships.append(lms.model.groups.GroupMembership(group = group_query, user = user_query))

    return memberships



import typing

import edq.util.time

import lms.model.assignments
import lms.model.scores
import lms.model.testdata.assignments
import lms.model.testdata.courses
import lms.model.testdata.users
import lms.model.users

# {course_name: {assignment_name: {user_name: score, ...}, ...}
COURSE_ASSIGNMENT_SCORES: typing.Dict[str, typing.Dict[str, typing.Dict[str, lms.model.scores.AssignmentScore]]] = {}

COURSE_ASSIGNMENT_SCORES['Course 101'] = {
    'Homework 0': {
        'course-student': lms.model.scores.AssignmentScore(
            id = '110050101',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Course 101']['Homework 0'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Course 101']['course-student'].to_query(),
            score = 2.0,
            graded_date = edq.util.time.Timestamp(1697406273000),
        ),
    },
}

COURSE_ASSIGNMENT_SCORES['Extra Course'] = {
    'Assignment 1': {
        'extra-course-student-1': lms.model.scores.AssignmentScore(
            id = '130060101',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 1'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-1'].to_query(),
            score = 10.0,
            graded_date = edq.util.time.Timestamp(1100000000000),
        ),
        'extra-course-student-2': lms.model.scores.AssignmentScore(
            id = '130070101',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 1'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-2'].to_query(),
            score = 7.5,
            graded_date = edq.util.time.Timestamp(1100000000000),
        ),
        'extra-course-student-3': lms.model.scores.AssignmentScore(
            id = '130080101',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 1'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-3'].to_query(),
            score = 5.0,
            graded_date = edq.util.time.Timestamp(1100000000000),
        ),
        'extra-course-student-4': lms.model.scores.AssignmentScore(
            id = '130090101',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 1'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-4'].to_query(),
            score = 0.0,
            graded_date = edq.util.time.Timestamp(1100000000000),
        ),
    },
    'Assignment 2': {
        'extra-course-student-1': lms.model.scores.AssignmentScore(
            id = '130060201',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 2'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-1'].to_query(),
            score = 20.0,
            graded_date = edq.util.time.Timestamp(1200000000000),
        ),
        'extra-course-student-2': lms.model.scores.AssignmentScore(
            id = '130070201',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 2'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-2'].to_query(),
            score = 15.0,
            graded_date = edq.util.time.Timestamp(1200000000000),
        ),
        'extra-course-student-3': lms.model.scores.AssignmentScore(
            id = '130080201',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 2'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-3'].to_query(),
            score = 10.0,
            graded_date = edq.util.time.Timestamp(1200000000000),
        ),
        'extra-course-student-4': lms.model.scores.AssignmentScore(
            id = '130090201',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 2'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-4'].to_query(),
            score = 0.0,
            graded_date = edq.util.time.Timestamp(1200000000000),
        ),
    },
    'Assignment 3': {
        'extra-course-student-1': lms.model.scores.AssignmentScore(
            id = '130060301',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 3'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-1'].to_query(),
            score = 30.0,
            graded_date = edq.util.time.Timestamp(1300000000000),
        ),
        'extra-course-student-2': lms.model.scores.AssignmentScore(
            id = '130070301',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 3'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-2'].to_query(),
            score = 22.5,
            graded_date = edq.util.time.Timestamp(1300000000000),
        ),
        'extra-course-student-3': lms.model.scores.AssignmentScore(
            id = '130080301',
            assignment_query = lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 3'].to_query(),
            user_query = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-3'].to_query(),
            score = 15.0,
            graded_date = edq.util.time.Timestamp(1300000000000),
        ),
    },
}

# {course_name: {assignment_name: {user_name: score, ...}, ...}
COURSE_ASSIGNMENT_SCORES_UNRESOLVED: typing.Dict[str, typing.Dict[str, typing.Dict[str, lms.model.scores.AssignmentScore]]] = {}

# {course_name: gradebook}
COURSE_GRADEBOOKS: typing.Dict[str, lms.model.scores.Gradebook] = {}

# {course_name: gradebook}
COURSE_GRADEBOOKS_UNRESOLVED: typing.Dict[str, lms.model.scores.Gradebook] = {}

def _load_unresolved_scores() -> None:
    """ Load unresolved scores from resolved scores. """

    for (course_name, assignment_scores) in COURSE_ASSIGNMENT_SCORES.items():
        if (course_name not in COURSE_ASSIGNMENT_SCORES_UNRESOLVED):
            COURSE_ASSIGNMENT_SCORES_UNRESOLVED[course_name] = {}

        for (assignment_name, user_scores) in assignment_scores.items():
            if (assignment_name not in COURSE_ASSIGNMENT_SCORES_UNRESOLVED[course_name]):
                COURSE_ASSIGNMENT_SCORES_UNRESOLVED[course_name][assignment_name] = {}

            for (user_name, old_score) in user_scores.items():
                if ((old_score.assignment_query is None) or (old_score.user_query is None)):
                    raise ValueError(f"None query for {course_name}.{assignment_name}.{user_name}.")

                new_score = lms.model.scores.AssignmentScore(
                    id = old_score.id,
                    assignment_query = lms.model.assignments.AssignmentQuery(id = old_score.assignment_query.id),
                    user_query = lms.model.users.UserQuery(id = old_score.user_query.id),
                    score = old_score.score,
                    graded_date = old_score.graded_date,
                )

                COURSE_ASSIGNMENT_SCORES_UNRESOLVED[course_name][assignment_name][user_name] = new_score

def _load_gradebooks(
        gradebooks: typing.Dict[str, lms.model.scores.Gradebook],
        scores: typing.Dict[str, typing.Dict[str, typing.Dict[str, lms.model.scores.AssignmentScore]]],
        get_assignment_query: typing.Callable,
        get_user_query: typing.Callable,
        ) -> None:
    """ Load gradebooks from scores. """

    for course_name in lms.model.testdata.courses.COURSES.keys():
        assignments: typing.Any = lms.model.testdata.assignments.COURSE_ASSIGNMENTS[course_name].values()
        assignment_queries = list(sorted([get_assignment_query(assignment) for assignment in assignments]))

        users: typing.Any = lms.model.testdata.users.COURSE_USERS[course_name].values()
        users = list(filter(lambda user: user.is_student(), users))
        user_queries = list(sorted([get_user_query(user) for user in users]))

        gradebook = lms.model.scores.Gradebook(assignment_queries, user_queries)

        for assignments_users_scores in scores.get(course_name, {}).values():
            for score in assignments_users_scores.values():
                gradebook.add(score)

        gradebooks[course_name] = gradebook

_load_unresolved_scores()

_load_gradebooks(
    COURSE_GRADEBOOKS,
    COURSE_ASSIGNMENT_SCORES,
    lambda assignment: assignment.to_query(),
    lambda user: user.to_query(),
)

_load_gradebooks(
    COURSE_GRADEBOOKS_UNRESOLVED,
    COURSE_ASSIGNMENT_SCORES_UNRESOLVED,
    lambda assignment: lms.model.assignments.AssignmentQuery(id = assignment.id),
    lambda user: lms.model.users.UserQuery(id = user.id),
)



import argparse
import os
import typing

import edq.procedure.verify_exchanges
import edq.testing.run

import lms.backend.testing
import lms.backend.canvas.model
import lms.model.backend
import lms.testing.serverrunner

THIS_DIR: str = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
ROOT_PACKAGE_DIR: str = os.path.join(THIS_DIR, '..')

def generate(args: typing.Union[argparse.Namespace, typing.Dict[str, typing.Any]]) -> int:
    """
    Generate HTTP test data by
    setting an exchange path, pointing to a live LMS server, and running all tests.
    The arguments may come directly from the parser for lms.cli.lib.generate-test-data.
    """

    if (not isinstance(args, dict)):
        args = vars(args)

    args.update(args.get('_config', {}))

    server_runner = lms.testing.serverrunner.LMSServerRunner(**args)
    server_runner.start()

    # Configure backend tests.
    lms.backend.testing.BackendTest.allowed_backend = server_runner.backend_type
    lms.backend.testing.BackendTest.skip_test_exchanges_base = True
    lms.backend.testing.BackendTest.override_server_url = server_runner.server
    lms.model.backend.APIBackend._testing_override = False
    lms.backend.canvas.model._testing_override = True

    # Run the tests (which generate the data).
    test_args = {
        'test_dirs': [ROOT_PACKAGE_DIR],
        'fail_fast': args.get('fail_fast', False),
    }
    failure_count = int(edq.testing.run.run(test_args))

    server_runner.stop()

    return failure_count

def verify(args: typing.Union[argparse.Namespace, typing.Dict[str, typing.Any]]) -> int:
    """
    Verify that test data matches data returned by the LMS server.
    The arguments may come directly from the parser for lms.cli.lib.verify-test-data.
    """

    if (not isinstance(args, dict)):
        args = vars(args)

    args.update(args.get('_config', {}))

    test_data_dir = args.get('test_data_dir', None)
    if (test_data_dir is None):
        raise ValueError("No test data dir was providded.")

    server_runner = lms.testing.serverrunner.LMSServerRunner(**args)
    server_runner.start()

    failure_count = int(edq.procedure.verify_exchanges.run([test_data_dir], server_runner.server, fail_fast = args.get('fail_fast', False)))

    server_runner.stop()

    return failure_count



"""
Utilities for network and HTTP.
"""

import typing
import urllib.parse

import edq.util.json
import requests

import lms.model.constants

CANVAS_CLEAN_REMOVE_CONTENT_KEYS: typing.List[str] = [
    'created_at',
    'ics',
    'last_activity_at',
    'lti_context_id',
    'preview_url',
    'secure_params',
    'total_activity_time',
    'updated_at',
    'url',
    'uuid',
]
""" Keys to remove from Canvas content. """

BLACKBOARD_CLEAN_REMOVE_CONTENT_KEYS: typing.List[str] = [
    'created',
    'modified',
]
""" Keys to remove from Blackboard content. """

BLACKBOARD_CLEAN_REMOVE_HEADERS: typing.Set[str] = {
    'access-control-allow-origin',
    'content-encoding',
    'content-language',
    'expires',
    'last-modified',
    'p3p',
    'pragma',
    'strict-transport-security',
    'transfer-encoding',
    'vary',
    'x-blackboard-xsrf',
}
""" Keys to remove from Blackboard content. """

def clean_lms_response(response: requests.Response, body: str) -> str:
    """
    A ResponseModifierFunction that attempt to identify
    if the requests comes from a Learning Management System (LMS),
    and clean the response accordingly.
    """

    # Check the standard LMS Toolkit backend header.
    backend_type = response.headers.get(lms.model.constants.HEADER_KEY_BACKEND, '').lower()

    if (backend_type == lms.model.constants.BACKEND_TYPE_CANVAS):
        return clean_canvas_response(response, body)

    if (backend_type == lms.model.constants.BACKEND_TYPE_MOODLE):
        return clean_moodle_response(response, body)

    # Try looking inside the header keys.
    for key in response.headers:
        key = key.lower().strip()

        if ('blackboard' in key):
            return clean_blackboard_response(response, body)

        if ('canvas' in key):
            return clean_canvas_response(response, body)

        if ('moodle' in key):
            return clean_moodle_response(response, body)

    return body

def clean_blackboard_response(response: requests.Response, body: str) -> str:
    """
    See clean_lms_response(), but specifically for the Blackboard LMS.
    This function will:
     - Call _clean_base_response().
     - Remove specific headers.
    """

    body = _clean_base_response(response, body)

    # Work on both request and response headers.
    for headers in [response.headers, response.request.headers]:
        for key in list(headers.keys()):
            if (key.strip().lower() in BLACKBOARD_CLEAN_REMOVE_HEADERS):
                headers.pop(key, None)

    # Most blackboard responses are JSON.
    try:
        data = edq.util.json.loads(body, strict = True)
    except Exception:
        # Response is not JSON.
        return body

    # Remove any content keys.
    _recursive_remove_keys(data, set(BLACKBOARD_CLEAN_REMOVE_CONTENT_KEYS))

    # Convert body back to a string.
    body = edq.util.json.dumps(data)

    return body

def clean_canvas_response(response: requests.Response, body: str) -> str:
    """
    See clean_lms_response(), but specifically for the Canvas LMS.
    This function will:
     - Call _clean_base_response().
     - Remove content keys: [last_activity_at, total_activity_time]
    """

    body = _clean_base_response(response, body)

    # Most canvas responses are JSON.
    try:
        data = edq.util.json.loads(body, strict = True)
    except Exception:
        # Response is not JSON.
        return body

    # Remove any content keys.
    _recursive_remove_keys(data, set(CANVAS_CLEAN_REMOVE_CONTENT_KEYS))

    # Remove special fields.

    if ('submissions/update_grades' in response.request.url):
        data.pop('id', None)

    # Convert body back to a string.
    body = edq.util.json.dumps(data)

    return body

def clean_moodle_response(response: requests.Response, body: str) -> str:
    """
    See clean_lms_response(), but specifically for the Moodle LMS.
    This function will:
     - Call _clean_base_response().
    """

    body = _clean_base_response(response, body)

    return body

def _clean_base_response(response: requests.Response, body: str,
        keep_headers: typing.Union[typing.List[str], None] = None) -> str:
    """
    Do response cleaning that is common amongst all backend types.
    This function will:
     - Remove X- headers.
    """

    # Index requests are generally for identification, and we use headers.
    path = urllib.parse.urlparse(response.request.url).path.strip()
    if (path in ['', '/']):
        body = ''

    for key in list(response.headers.keys()):
        key = key.strip().lower()
        if ((keep_headers is not None) and (key in keep_headers)):
            continue

        if (key.startswith('x-')):
            response.headers.pop(key, None)

    return body

def _recursive_remove_keys(data: typing.Any, remove_keys: typing.Set[str]) -> None:
    """
    Recursively descend through the given and remove any instance to the given key from any dictionaries.
    The data should only be simple types (POD, dicts, lists, tuples).
    """

    if (isinstance(data, (list, tuple))):
        for item in data:
            _recursive_remove_keys(item, remove_keys)
    elif (isinstance(data, dict)):
        for key in list(data.keys()):
            if (key in remove_keys):
                del data[key]
            else:
                _recursive_remove_keys(data[key], remove_keys)

def parse_cookies(
        text_cookies: typing.Union[str, None],
        strip_key_prefix: bool = True,
        ) -> typing.Dict[str, typing.Any]:
    """ Parse cookies out of a text string. """

    cookies: typing.Dict[str, typing.Any] = {}

    if (text_cookies is None):
        return cookies

    text_cookies = text_cookies.strip()
    if (len(text_cookies) == 0):
        return cookies

    for cookie in text_cookies.split('; '):
        parts = cookie.split('=', maxsplit = 1)

        key = parts[0].lower()

        if (strip_key_prefix):
            key = key.split(', ')[-1]

        if (len(parts) == 1):
            cookies[key] = True
        else:
            cookies[key] = parts[1]

    return cookies



import lms.backend.testing
import lms.model.testdata.scores

def test_courses_users_scores_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of listing users' scores. """

    scores = lms.model.testdata.scores.COURSE_ASSIGNMENT_SCORES_UNRESOLVED

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_id': '110000000',
                'user_id': '100050000',
            },
            [
                scores['Course 101']['Homework 0']['course-student'],
            ],
            None,
        ),

        (
            {
                'course_id': '120000000',
                'user_id': '100050000',
            },
            [
            ],
            None,
        ),

        (
            {
                'course_id': '130000000',
                'user_id': '100060000',
            },
            [
                scores['Extra Course']['Assignment 1']['extra-course-student-1'],
                scores['Extra Course']['Assignment 2']['extra-course-student-1'],
                scores['Extra Course']['Assignment 3']['extra-course-student-1'],
            ],
            None,
        ),

        (
            {
                'course_id': '130000000',
                'user_id': '100070000',
            },
            [
                scores['Extra Course']['Assignment 1']['extra-course-student-2'],
                scores['Extra Course']['Assignment 2']['extra-course-student-2'],
                scores['Extra Course']['Assignment 3']['extra-course-student-2'],
            ],
            None,
        ),

        (
            {
                'course_id': '130000000',
                'user_id': '100080000',
            },
            [
                scores['Extra Course']['Assignment 1']['extra-course-student-3'],
                scores['Extra Course']['Assignment 2']['extra-course-student-3'],
                scores['Extra Course']['Assignment 3']['extra-course-student-3'],
            ],
            None,
        ),

        (
            {
                'course_id': '130000000',
                'user_id': '100090000',
            },
            [
                scores['Extra Course']['Assignment 1']['extra-course-student-4'],
                scores['Extra Course']['Assignment 2']['extra-course-student-4'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_users_scores_list, test_cases)



import typing

import lms.model.groups
import lms.model.groupsets
import lms.model.testdata.groupsets
import lms.model.testdata.users

# {course_name: {name: group, ...}, ...}
COURSE_GROUPS: typing.Dict[str, typing.Dict[str, lms.model.groups.Group]] = {}

COURSE_GROUPS['Extra Course'] = {
    'Group 1-1': lms.model.groups.Group(
        id = '131010101',
        name = 'Group 1-1',
    ),
    'Group 1-2': lms.model.groups.Group(
        id = '131010102',
        name = 'Group 1-2',
    ),
    'Group 2-1': lms.model.groups.Group(
        id = '131020201',
        name = 'Group 2-1',
    ),
    'Group 2-2': lms.model.groups.Group(
        id = '131020202',
        name = 'Group 2-2',
    ),
    'Group 3-1': lms.model.groups.Group(
        id = '131030301',
        name = 'Group 3-1',
    ),
}

# {course_name: {groupset_name: {user name: membership, ...}, ...}, ...}
COURSE_GROUP_MEMBERSHIPS: typing.Dict[str, typing.Dict[str, typing.Dict[str, lms.model.groupsets.GroupSetMembership]]] = {}

COURSE_GROUP_MEMBERSHIPS['Extra Course'] = {
    'Group Set 1': {
        'extra-course-student-1': lms.model.groupsets.GroupSetMembership(
            groupset = lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'].to_query(),
            group = COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(),
            user = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-1'].to_query(),
        ),
        'extra-course-student-2': lms.model.groupsets.GroupSetMembership(
            groupset = lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'].to_query(),
            group = COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(),
            user = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-2'].to_query(),
        ),
        'extra-course-student-3': lms.model.groupsets.GroupSetMembership(
            groupset = lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'].to_query(),
            group = COURSE_GROUPS['Extra Course']['Group 1-2'].to_query(),
            user = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-3'].to_query(),
        ),
        'extra-course-student-4': lms.model.groupsets.GroupSetMembership(
            groupset = lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'].to_query(),
            group = COURSE_GROUPS['Extra Course']['Group 1-2'].to_query(),
            user = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-4'].to_query(),
        ),
    },
    'Group Set 2': {
        'extra-course-student-1': lms.model.groupsets.GroupSetMembership(
            groupset = lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 2'].to_query(),
            group = COURSE_GROUPS['Extra Course']['Group 2-1'].to_query(),
            user = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-1'].to_query(),
        ),
        'extra-course-student-2': lms.model.groupsets.GroupSetMembership(
            groupset = lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 2'].to_query(),
            group = COURSE_GROUPS['Extra Course']['Group 2-2'].to_query(),
            user = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-2'].to_query(),
        ),
        'extra-course-student-3': lms.model.groupsets.GroupSetMembership(
            groupset = lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 2'].to_query(),
            group = COURSE_GROUPS['Extra Course']['Group 2-1'].to_query(),
            user = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-3'].to_query(),
        ),
        'extra-course-student-4': lms.model.groupsets.GroupSetMembership(
            groupset = lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 2'].to_query(),
            group = COURSE_GROUPS['Extra Course']['Group 2-2'].to_query(),
            user = lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-4'].to_query(),
        ),
    },
}



"""
Set membership for a group set.
"""

import argparse
import sys

import lms.backend.instance
import lms.cli.common
import lms.cli.courses.groupsets.memberships.common
import lms.cli.parser
import lms.model.base

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    config = args._config

    backend = lms.backend.instance.get_backend(**config)

    course_query = lms.cli.common.check_required_course(backend, config)
    if (course_query is None):
        return 1

    groupset_query = lms.cli.common.check_required_groupset(backend, config)
    if (groupset_query is None):
        return 2

    memberships = lms.cli.courses.groupsets.memberships.common.load_group_memberships(backend, args.path, args.skip_rows)

    stats = backend.courses_groupsets_memberships_resolve_and_set(course_query, groupset_query, memberships)
    (created_groups, deleted_groups, add_counts, sub_counts) = stats

    if (len(created_groups) > 0):
        display_groups = [str(group.to_query()) for group in created_groups]
        print(f"Created {len(created_groups)} groups: {display_groups}.")

    if (len(deleted_groups) > 0):
        display_groups = [str(group_query) for group_query in deleted_groups]
        print(f"Deleted {len(deleted_groups)} groups: {display_groups}.")

    for group_query in sorted(set(add_counts.keys()) | set(sub_counts.keys())):
        add_count = add_counts.get(group_query, 0)
        sub_count = sub_counts.get(group_query, 0)

        print(f"Added {add_count} and subtracted {sub_count} users to/from group {group_query}.")

    return 0

def main() -> int:
    """ Get a parser, parse the args, and call run. """
    return run_cli(_get_parser().parse_args())

def _get_parser() -> argparse.ArgumentParser:
    """ Get the parser. """

    parser = lms.cli.parser.get_parser(__doc__.strip(),
            include_output_format = True,
            include_course = True,
            include_groupset = True,
            include_group = True,
            include_skip_rows = True,
    )

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to a TSV file where each row has 2 columns: group query and user query.')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import lms.backend.testing
import lms.model.users
import lms.model.testdata.users

def test_courses_users_get_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of getting course users. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Empty
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [],
            },
            [
            ],
            None,
        ),

        # Base - List
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(id = '100010000'),
                    lms.model.users.UserQuery(id = '100020000'),
                ],
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-grader'],
            ],
            None,
        ),

        # Base - Fetch
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(id = '100010000'),
                ],
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
            ],
            None,
        ),

        # Course Query
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Course 101'),
                'user_queries': [
                    lms.model.users.UserQuery(id = '100010000'),
                ],
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
            ],
            None,
        ),

        # Query - Name
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(name = 'course-admin'),
                ],
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
            ],
            None,
        ),

        # Query - Email
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(email = 'course-admin@test.edulinq.org'),
                ],
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
            ],
            None,
        ),

        # Query - Label Name
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(name = 'course-admin', id = '100010000'),
                ],
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
            ],
            None,
        ),

        # Query - Label Email
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(email = 'course-admin@test.edulinq.org', id = '100010000'),
                ],
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
            ],
            None,
        ),

        # Miss - Course
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '999'),
                'user_queries': [
                    lms.model.users.UserQuery(id = '100010000'),
                ],
            },
            None,
            'Could not resolve course query',
        ),

        # Miss - ID
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(id = '999'),
                ],
            },
            [
            ],
            None,
        ),

        # Miss - Name
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(name = 'ZZZ'),
                ],
            },
            [
            ],
            None,
        ),

        # Miss - Email
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(email = 'ZZZ@test.edulinq.org'),
                ],
            },
            [
            ],
            None,
        ),

        # Miss - Partial Match
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(id = '100010000', name = 'ZZZ'),
                ],
            },
            [
            ],
            None,
        ),

        # Multiple Match
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_queries': [
                    lms.model.users.UserQuery(id = '100010000'),
                    lms.model.users.UserQuery(name = 'course-admin'),
                ],
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_users_get, test_cases)



import typing

import lms.model.users

# {course_name: {user_name: user, ...}, ...}
COURSE_USERS: typing.Dict[str, typing.Dict[str, lms.model.users.CourseUser]] = {}

COURSE_USERS['Course 101'] = {
    'course-admin': lms.model.users.CourseUser(
        email = 'course-admin@test.edulinq.org',
        id = '100010000',
        name = 'course-admin',
        role = lms.model.users.CourseRole.ADMIN,
    ),
    'course-grader': lms.model.users.CourseUser(
        email = 'course-grader@test.edulinq.org',
        id = '100020000',
        name = 'course-grader',
        role = lms.model.users.CourseRole.GRADER,
    ),
    'course-other': lms.model.users.CourseUser(
        email = 'course-other@test.edulinq.org',
        id = '100030000',
        name = 'course-other',
        role = lms.model.users.CourseRole.OTHER,
    ),
    'course-owner': lms.model.users.CourseUser(
        email = 'course-owner@test.edulinq.org',
        id = '100040000',
        name = 'course-owner',
        role = lms.model.users.CourseRole.OWNER,
    ),
    'course-student': lms.model.users.CourseUser(
        email = 'course-student@test.edulinq.org',
        id = '100050000',
        name = 'course-student',
        role = lms.model.users.CourseRole.STUDENT,
    ),
}

COURSE_USERS['Course Using Different Languages'] = COURSE_USERS['Course 101']

COURSE_USERS['Extra Course'] = {
    'course-owner': COURSE_USERS['Course 101']['course-owner'],
    'extra-course-student-1': lms.model.users.CourseUser(
        email = 'extra-course-student-1@test.edulinq.org',
        id = '100060000',
        name = 'extra-course-student-1',
        role = lms.model.users.CourseRole.STUDENT,
    ),
    'extra-course-student-2': lms.model.users.CourseUser(
        email = 'extra-course-student-2@test.edulinq.org',
        id = '100070000',
        name = 'extra-course-student-2',
        role = lms.model.users.CourseRole.STUDENT,
    ),
    'extra-course-student-3': lms.model.users.CourseUser(
        email = 'extra-course-student-3@test.edulinq.org',
        id = '100080000',
        name = 'extra-course-student-3',
        role = lms.model.users.CourseRole.STUDENT,
    ),
    'extra-course-student-4': lms.model.users.CourseUser(
        email = 'extra-course-student-4@test.edulinq.org',
        id = '100090000',
        name = 'extra-course-student-4',
        role = lms.model.users.CourseRole.STUDENT,
    ),
}

# {user_name: user, ...}
SERVER_USERS = {
    'course-admin': lms.model.users.ServerUser(
        email = 'course-admin@test.edulinq.org',
        id = '100010000',
        name = 'course-admin',
    ),
    'course-grader': lms.model.users.ServerUser(
        email = 'course-grader@test.edulinq.org',
        id = '100020000',
        name = 'course-grader',
    ),
    'course-other': lms.model.users.ServerUser(
        email = 'course-other@test.edulinq.org',
        id = '100030000',
        name = 'course-other',
    ),
    'course-owner': lms.model.users.ServerUser(
        email = 'course-owner@test.edulinq.org',
        id = '100040000',
        name = 'course-owner',
    ),
    'course-student': lms.model.users.ServerUser(
        email = 'course-student@test.edulinq.org',
        id = '100050000',
        name = 'course-student',
    ),
    'extra-course-student-1': lms.model.users.ServerUser(
        email = 'extra-course-student-1@test.edulinq.org',
        id = '100060000',
        name = 'extra-course-student-1',
    ),
    'extra-course-student-2': lms.model.users.ServerUser(
        email = 'extra-course-student-2@test.edulinq.org',
        id = '100070000',
        name = 'extra-course-student-2',
    ),
    'extra-course-student-3': lms.model.users.ServerUser(
        email = 'extra-course-student-3@test.edulinq.org',
        id = '100080000',
        name = 'extra-course-student-3',
    ),
    'extra-course-student-4': lms.model.users.ServerUser(
        email = 'extra-course-student-4@test.edulinq.org',
        id = '100090000',
        name = 'extra-course-student-4',
    ),
    'sevrer-admin': lms.model.users.ServerUser(
        email = 'server-admin@test.edulinq.org',
        id = '100100000',
        name = 'server-admin',
    ),
    'sevrer-creator': lms.model.users.ServerUser(
        email = 'server-creator@test.edulinq.org',
        id = '100110000',
        name = 'server-creator',
    ),
    'sevrer-owner': lms.model.users.ServerUser(
        email = 'server-owner@test.edulinq.org',
        id = '100120000',
        name = 'server-owner',
    ),
    'sevrer-user': lms.model.users.ServerUser(
        email = 'server-user@test.edulinq.org',
        id = '100130000',
        name = 'server-user',
    ),
}



import re
import typing

import edq.testing.asserts
import edq.testing.unittest

ID_SUBS: typing.List[typing.Tuple[re.Pattern, str]] = [
    (re.compile(r'^id: \d+'), 'id: <ID>'),  # Text
    (re.compile(r'^\d+\t'), "<ID>\t"),  # Table
    (re.compile(r'"id": "\d+"'), '"id": "<ID>"'),  # JSON
    (re.compile(r' \(\d{1,2}\)'), ' (<ID>)'),  # Labels with Small IDs
]

def cli_assert_normalize_ids(test: edq.testing.unittest.BaseTest, expected: str, actual: str, **kwargs: typing.Any) -> None:
    """
    Normalize IDs (as output from a standard CLI,
    and then use edq.testing.asserts.content_equals_normalize().
    """

    for (pattern, replacement) in ID_SUBS:
        expected = re.sub(pattern, replacement, expected)
        actual = re.sub(pattern, replacement, actual)

    edq.testing.asserts.content_equals_normalize(test, expected, actual)



import edq.testing.unittest

import lms.model.base
import lms.model.courses
import lms.model.users
import lms.model.assignments

class TestBaseType(edq.testing.unittest.BaseTest):
    """ Test functonality of types derived from the base type. """

    # JSON Dict

    def test_as_json_dict_base(self):
        """ Test converting to a JSON dict. """

        # [(input, kwargs, expected), ...]
        test_cases = [
            # Base
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {},
                {
                    'email': None,
                    'id': '123',
                    'name': 'Alice',
                },
            ),
            (
                lms.model.assignments.Assignment(id = '123', name = 'Assignment 1', extra = 'abc'),
                {},
                {
                    'id': '123',
                    'name': 'Assignment 1',
                    'close_date': None,
                    'description': None,
                    'due_date': None,
                    'open_date': None,
                    'points_possible': None,
                },
            ),
            (
                lms.model.courses.Course(id = '123', name = 'Course 1', extra = 'abc'),
                {},
                {
                    'id': '123',
                    'name': 'Course 1',
                },
            ),

            # Extra Fields
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {
                    'include_extra_fields': True,
                },
                {
                    'email': None,
                    'id': '123',
                    'name': 'Alice',
                    'extra': 'abc',
                },
            ),
        ]

        for (i, test_case) in enumerate(test_cases):
            (value, kwargs, expected) = test_case

            with self.subTest(msg = f"Case {i} ('{value}'):"):
                actual = value.as_json_dict(**kwargs)
                self.assertJSONDictEqual(expected, actual)

    def test_from_json_dict_base(self):
        """ Test converting from a JSON dict. """

        # [(input, kwargs, expected), ...]
        test_cases = [
            # Base - User
            (
                {
                    'email': None,
                    'id': '123',
                    'name': 'Alice',
                },
                {},
                lms.model.users.ServerUser(id = '123', name = 'Alice'),
            ),
            (
                {
                    'id': '123',
                    'name': 'Assignment 1',
                },
                {},
                lms.model.assignments.Assignment(id = '123', name = 'Assignment 1'),
            ),
            (
                {
                    'id': '123',
                    'name': 'Course 1',
                },
                {},
                lms.model.courses.Course(id = '123', name = 'Course 1'),
            ),
        ]

        for (i, test_case) in enumerate(test_cases):
            (value, kwargs, expected) = test_case

            with self.subTest(msg = f"Case {i} ('{value}'):"):
                actual = type(expected).from_json_dict(value, **kwargs)
                self.assertJSONEqual(expected, actual)

    def test_json_dict_cyclic_serialization(self):
        """ Test converting to a JSON dict and back to an object. """

        # [(input, kwargs), ...]
        test_cases = [
            # Base
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {},
            ),
            (
                lms.model.assignments.Assignment(id = '123', name = 'Assignment 1', extra = 'abc'),
                {},
            ),
            (
                lms.model.courses.Course(id = '123', name = 'Course 1', extra = 'abc'),
                {},
            ),

            # Extra Fields
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {
                    'include_extra_fields': True,
                },
            ),
        ]

        for (i, test_case) in enumerate(test_cases):
            (value, kwargs) = test_case

            with self.subTest(msg = f"Case {i} ('{value}'):"):
                json_dict = value.as_json_dict(**kwargs)
                new_value = type(value).from_json_dict(json_dict, **kwargs)
                self.assertJSONEqual(value, new_value)

    # Text Rows

    def test_as_text_rows_base(self):
        """ Test converting to text rows. """

        # [(input, kwargs, expected), ...]
        test_cases = [
            # Base
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {},
                [
                    'id: 123',
                    'name: Alice',
                    'email: ',
                ],
            ),
            (
                lms.model.assignments.Assignment(id = '123', name = 'Assignment 1', extra = 'abc'),
                {},
                [
                    'id: 123',
                    'name: Assignment 1',
                    'description: ',
                    'open_date: ',
                    'close_date: ',
                    'due_date: ',
                    'points_possible: ',
                ],
            ),
            (
                lms.model.courses.Course(id = '123', name = 'Course 1', extra = 'abc'),
                {},
                [
                    'id: 123',
                    'name: Course 1',
                ],
            ),

            # Extra Fields
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {
                    'include_extra_fields': True,
                },
                [
                    'id: 123',
                    'name: Alice',
                    'email: ',
                    'extra: abc',
                ],
            ),

            # Skip Headers
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {
                    'skip_headers': True,
                },
                [
                    '123',
                    'Alice',
                    '',
                ],
            ),

            # Pretty Headers
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {
                    'pretty_headers': True,
                },
                [
                    'Id: 123',
                    'Name: Alice',
                    'Email: ',
                ],
            ),

            # Skip Pretty Headers
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {
                    'skip_headers': True,
                    'pretty_headers': True,
                },
                [
                    '123',
                    'Alice',
                    '',
                ],
            ),
        ]

        for (i, test_case) in enumerate(test_cases):
            (value, kwargs, expected) = test_case

            with self.subTest(msg = f"Case {i} ('{value}'):"):
                actual = value.as_text_rows(**kwargs)
                self.assertJSONListEqual(expected, actual)

    # Table Rows

    def test_as_table_rows_base(self):
        """ Test converting to table rows. """

        # [(input, kwargs, expected), ...]
        test_cases = [
            # Base
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {},
                [[
                    '123',
                    'Alice',
                    '',
                ]],
            ),
            (
                lms.model.assignments.Assignment(id = '123', name = 'Assignment 1', extra = 'abc'),
                {},
                [[
                    '123',
                    'Assignment 1',
                    '',
                    '',
                    '',
                    '',
                    '',
                ]],
            ),
            (
                lms.model.courses.Course(id = '123', name = 'Course 1', extra = 'abc'),
                {},
                [[
                    '123',
                    'Course 1',
                ]],
            ),

            # Extra Fields
            (
                lms.model.users.ServerUser(id = '123', name = 'Alice', extra = 'abc'),
                {
                    'include_extra_fields': True,
                },
                [[
                    '123',
                    'Alice',
                    '',
                    'abc',
                ]],
            ),
        ]

        for (i, test_case) in enumerate(test_cases):
            (value, kwargs, expected) = test_case

            with self.subTest(msg = f"Case {i} ('{value}'):"):
                actual = value.as_table_rows(**kwargs)
                self.assertJSONEqual(expected, actual)



import typing

import lms.model.groupsets

# {course_name: {name: groupset, ...}, ...}
COURSE_GROUPSETS: typing.Dict[str, typing.Dict[str, lms.model.groupsets.GroupSet]] = {}

COURSE_GROUPSETS['Extra Course'] = {
    'Group Set 1': lms.model.groupsets.GroupSet(
        id = '131010100',
        name = 'Group Set 1',
    ),
    'Group Set 2': lms.model.groupsets.GroupSet(
        id = '131020200',
        name = 'Group Set 2',
    ),
    'Group Set 3': lms.model.groupsets.GroupSet(
        id = '131030300',
        name = 'Group Set 3',
    ),
}



import typing

import lms.model.assignments

# {course_name: {name: assignment, ...}, ...}
COURSE_ASSIGNMENTS: typing.Dict[str, typing.Dict[str, lms.model.assignments.Assignment]] = {}

COURSE_ASSIGNMENTS['Course 101'] = {
    'Homework 0': lms.model.assignments.Assignment(
        id = '110000100',
        name = 'Homework 0',
        points_possible = 2.0,
    ),
}

COURSE_ASSIGNMENTS['Course Using Different Languages'] = {
    'A Simple Bash Assignment': lms.model.assignments.Assignment(
        id = '120000100',
        name = 'A Simple Bash Assignment',
        points_possible = 10.0,
    ),
    'A Simple C++ Assignment': lms.model.assignments.Assignment(
        id = '120000200',
        name = 'A Simple C++ Assignment',
        points_possible = 10.0,
    ),
    'A Simple Java Assignment': lms.model.assignments.Assignment(
        id = '120000300',
        name = 'A Simple Java Assignment',
        points_possible = 10.0,
    ),
}

COURSE_ASSIGNMENTS['Extra Course'] = {
    'Assignment 1': lms.model.assignments.Assignment(
        id = '130000100',
        name = 'Assignment 1',
        points_possible = 10.0,
    ),
    'Assignment 2': lms.model.assignments.Assignment(
        id = '130000200',
        name = 'Assignment 2',
        points_possible = 20.0,
    ),
    'Assignment 3': lms.model.assignments.Assignment(
        id = '130000300',
        name = 'Assignment 3',
        points_possible = 30.0,
    ),
}



import lms.backend.testing
import lms.model.testdata.groups

def test_courses_groups_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of listing course groups. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_id': '110000000',
                'groupset_id': '999',
            },
            [
            ],
            None,
        ),
        (
            {
                'course_id': '120000000',
                'groupset_id': '999',
            },
            [
            ],
            None,
        ),

        (
            {
                'course_id': '130000000',
                'groupset_id': '131010100',
            },
            [
                lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-1'],
                lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-2'],
            ],
            None,
        ),
        (
            {
                'course_id': '130000000',
                'groupset_id': '131020200',
            },
            [
                lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 2-1'],
                lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 2-2'],
            ],
            None,
        ),
        (
            {
                'course_id': '130000000',
                'groupset_id': '131030300',
            },
            [
                lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 3-1'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_groups_list, test_cases)



import lms.backend.testing
import lms.model.groupsets
import lms.model.testdata.groupsets

def test_courses_groupsets_get_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of getting course groupsets. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Empty
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'groupset_queries': [],
            },
            [
            ],
            None,
        ),

        # Course Query
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_queries': [
                    lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                ],
            },
            [
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'],
            ],
            None,
        ),

        # Miss - Course Query
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'ZZZ'),
                'groupset_queries': [
                    lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                ],
            },
            None,
            'Could not resolve course query',
        ),

        # Single
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'groupset_queries': [
                    lms.model.groupsets.GroupSetQuery(id = '131010100'),
                ],
            },
            [
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'],
            ],
            None,
        ),

        # Multiple
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'groupset_queries': [
                    lms.model.groupsets.GroupSetQuery(id = '131010100'),
                    lms.model.groupsets.GroupSetQuery(id = '131020200'),
                ],
            },
            [
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'],
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 2'],
            ],
            None,
        ),

        # Query - Name
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'groupset_queries': [
                    lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                ],
            },
            [
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'],
            ],
            None,
        ),

        # Query - Label
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'groupset_queries': [
                    lms.model.groupsets.GroupSetQuery(name = 'Group Set 2', id = '131020200'),
                ],
            },
            [
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 2'],
            ],
            None,
        ),

        # Miss - ID
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'groupset_queries': [
                    lms.model.groupsets.GroupSetQuery(id = 999),
                ],
            },
            [
            ],
            None,
        ),

        # Miss - Query
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'groupset_queries': [
                    lms.model.groupsets.GroupSetQuery(name = 'ZZZ'),
                ],
            },
            [
            ],
            None,
        ),

        # Miss - Partial Match
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'groupset_queries': [
                    lms.model.groupsets.GroupSetQuery(id = '131010100', name = 'ZZZ'),
                ],
            },
            [
            ],
            None,
        ),

        # Multiple Match
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'groupset_queries': [
                    lms.model.groupsets.GroupSetQuery(id = '131010100'),
                    lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                ],
            },
            [
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_groupsets_get, test_cases)



import lms.backend.testing
import lms.model.testdata.assignments

def test_courses_assignments_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of listing course assignments. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_id': '110000000',
            },
            [
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Course 101']['Homework 0'],
            ],
            None,
        ),
        (
            {
                'course_id': '120000000',
            },
            [
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Course Using Different Languages']['A Simple Bash Assignment'],
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Course Using Different Languages']['A Simple C++ Assignment'],
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Course Using Different Languages']['A Simple Java Assignment'],
            ],
            None,
        ),
        (
            {
                'course_id': '130000000',
            },
            [
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 1'],
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 2'],
                lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 3'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_assignments_list, test_cases)



import lms.backend.testing
import lms.model.courses
import lms.model.testdata.scores

def test_courses_users_scores_resolve_and_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of listing users' scores. """

    scores = lms.model.testdata.scores.COURSE_ASSIGNMENT_SCORES

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_query': lms.model.users.UserQuery(id = '100050000'),
            },
            [
                scores['Course 101']['Homework 0']['course-student'],
            ],
            None,
        ),

        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Course 101'),
                'user_query': lms.model.users.UserQuery(id = '100050000'),
            },
            [
                scores['Course 101']['Homework 0']['course-student'],
            ],
            None,
        ),

        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '120000000'),
                'user_query': lms.model.users.UserQuery(id = '100050000'),
            },
            [
            ],
            None,
        ),

        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'user_query': lms.model.users.UserQuery(id = '100060000'),
            },
            [
                scores['Extra Course']['Assignment 1']['extra-course-student-1'],
                scores['Extra Course']['Assignment 2']['extra-course-student-1'],
                scores['Extra Course']['Assignment 3']['extra-course-student-1'],
            ],
            None,
        ),

        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'user_query': lms.model.users.UserQuery(id = '100070000'),
            },
            [
                scores['Extra Course']['Assignment 1']['extra-course-student-2'],
                scores['Extra Course']['Assignment 2']['extra-course-student-2'],
                scores['Extra Course']['Assignment 3']['extra-course-student-2'],
            ],
            None,
        ),

        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'user_query': lms.model.users.UserQuery(id = '100080000'),
            },
            [
                scores['Extra Course']['Assignment 1']['extra-course-student-3'],
                scores['Extra Course']['Assignment 2']['extra-course-student-3'],
                scores['Extra Course']['Assignment 3']['extra-course-student-3'],
            ],
            None,
        ),

        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'user_query': lms.model.users.UserQuery(id = '100090000'),
            },
            [
                scores['Extra Course']['Assignment 1']['extra-course-student-4'],
                scores['Extra Course']['Assignment 2']['extra-course-student-4'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_users_scores_resolve_and_list, test_cases)



"""
Attempt to identify the backend of the target server.
"""

import argparse
import sys

import lms
import lms.backend.instance
import lms.cli.parser

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    config = args._config

    backend_type = lms.backend.instance.guess_backend_type(**config)
    if (backend_type is None):
        print(f"ERROR: Unable to determine backend type from server '{config.get('server', '')}'.")
        return 1

    print(backend_type)
    return 0

def main() -> int:
    """ Get a parser, parse the args, and call run. """
    return run_cli(_get_parser().parse_args())

def _get_parser() -> argparse.ArgumentParser:
    """ Get the parser. """

    return lms.cli.parser.get_parser(__doc__.strip())

if (__name__ == '__main__'):
    sys.exit(main())



import typing

import lms.backend.canvas.common
import lms.backend.canvas.model
import lms.model.backend
import lms.model.users

BASE_ENDPOINT = "/api/v1/courses/{course_id}/users?per_page={page_size}"

def request(backend: lms.model.backend.APIBackend,
        course_id: int,
        include_role: bool = True,
        ) -> typing.List[lms.model.users.CourseUser]:
    """ List course users. """

    url = backend.server + BASE_ENDPOINT.format(course_id = course_id, page_size = lms.backend.canvas.common.DEFAULT_PAGE_SIZE)
    headers = backend.get_standard_headers()

    if (include_role):
        url += '&include[]=enrollments'

    raw_objects = lms.backend.canvas.common.make_get_request_list(url, headers = headers)
    if (raw_objects is None):
        identifiers = {
            'course_id': course_id,
        }
        backend.not_found('list users', identifiers)

        return []

    return [lms.backend.canvas.model.course_user(backend, raw_object) for raw_object in raw_objects]



import typing

import lms.backend.canvas.common
import lms.backend.canvas.model
import lms.model.assignments
import lms.model.scores
import lms.model.users

BASE_ENDPOINT = "/api/v1/courses/{course_id}/students/submissions"

def request(backend: typing.Any,
        course_id: int,
        assignment_ids: typing.List[int],
        user_ids: typing.List[int],
        ) -> lms.model.scores.Gradebook:
    """ Fetch a gradebook for the given users and assignments. """

    url = backend.server + BASE_ENDPOINT.format(course_id = course_id)
    headers = backend.get_standard_headers()

    data = {
        'per_page': lms.backend.canvas.common.DEFAULT_PAGE_SIZE,
        'assignment_ids[]': [str(assignment_id) for assignment_id in assignment_ids],
        'student_ids[]': 'all',
    }

    assignment_queries = [lms.model.assignments.AssignmentQuery(id = id) for id in assignment_ids]
    user_queries = [lms.model.users.UserQuery(id = id) for id in user_ids]
    gradebook = lms.model.scores.Gradebook(assignment_queries, user_queries)

    raw_objects = lms.backend.canvas.common.make_get_request_list(url, headers = headers, data = data)
    if (raw_objects is None):
        identifiers = {
            'course_id': course_id,
            'assignment_ids': assignment_ids,
            'user_ids': user_ids,
        }
        backend.not_found('fetch gradebook', identifiers)

        return gradebook

    for raw_object in raw_objects:
        # Check if this is an actual submission and not just a placeholder.
        if (raw_object.get('workflow_state', None) == 'unsubmitted'):
            continue

        user_id = int(raw_object.get('user_id', -1))
        if ((len(user_ids) != 0) and (user_id not in user_ids)):
            continue

        gradebook.add(lms.backend.canvas.model.assignment_score(raw_object))

    return gradebook



import typing

def required_int(raw_value: typing.Any, name: str) -> int:
    """ Parse and clean an int value. """

    if (raw_value is None):
        raise ValueError(f"Value '{name}' is None, when it should be an int.")

    return int(raw_value)

def required_string(raw_value: typing.Any, name: str, **kwargs: typing.Any) -> str:
    """ Parse and clean a string value. """

    value = optional_string(raw_value, **kwargs)
    if (value is None):
        raise ValueError(f"Value '{name}' is None, when it should be a string.")

    return value

def optional_string(raw_value: typing.Any,
        strip: bool = True,
        allow_empty: bool = False,
        **kwargs: typing.Any) -> typing.Union[str, None]:
    """ Parse and clean an optional string value. """

    if (raw_value is None):
        return None

    value = str(raw_value)
    if (strip):
        value = value.strip()

    if ((not allow_empty) and (len(value) == 0)):
        return None

    return value

def optional_float(raw_value: typing.Any, label: str = 'field',
        **kwargs: typing.Any) -> typing.Union[float, None]:
    """ Parse and clean an optional float value. """

    if (raw_value is None):
        return None

    if (isinstance(raw_value, (int, float))):
        return float(raw_value)

    str_value = str(raw_value).strip()

    try:
        return float(str_value)
    except ValueError as ex:
        raise ValueError(f"Unable to parse {label} '{raw_value}' ({type(raw_value)}) as a float.") from ex



import typing

import lms.backend.canvas.common
import lms.backend.canvas.model
import lms.model.backend
import lms.model.groups
import lms.model.groupsets
import lms.model.users

BASE_ENDPOINT = "/api/v1/groups/{group_id}/users?per_page={page_size}"

def request(backend: lms.model.backend.APIBackend,
        course_id: int,
        groupset_id: int,
        group_id: int,
        ) -> typing.List[lms.model.groupsets.GroupSetMembership]:
    """ List a group memberships. """

    url = backend.server + BASE_ENDPOINT.format(group_id = group_id, page_size = lms.backend.canvas.common.DEFAULT_PAGE_SIZE)
    headers = backend.get_standard_headers()

    raw_objects = lms.backend.canvas.common.make_get_request_list(url, headers = headers)
    if (raw_objects is None):
        identifiers = {
            'course_id': course_id,
            'groupset_id': groupset_id,
            'group_id': group_id,
        }
        backend.not_found('list group memberships', identifiers)

        return []

    groupset_query = lms.model.groupsets.GroupSetQuery(id = groupset_id)
    group_query = lms.model.groups.GroupQuery(id = group_id)

    memberships = []
    for raw_object in raw_objects:
        user = lms.backend.canvas.model.course_user(backend, raw_object)
        membership = lms.model.groupsets.GroupSetMembership(user = user.to_query(), groupset = groupset_query, group = group_query)
        memberships.append(membership)

    return memberships



# pylint: disable=invalid-name

"""
Upload a single score (and optional comment) for an assignment.
"""

import argparse
import sys

import lms.backend.instance
import lms.cli.common
import lms.cli.parser
import lms.model.scores

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    config = args._config

    backend = lms.backend.instance.get_backend(**config)

    course_query = lms.cli.common.check_required_course(backend, config)
    if (course_query is None):
        return 1

    assignment_query = lms.cli.common.check_required_assignment(backend, config)
    if (assignment_query is None):
        return 2

    user_query = lms.cli.common.check_required_user(backend, config)
    if (user_query is None):
        return 3

    scores = {
        user_query: lms.model.scores.ScoreFragment(score = args.score, comment = args.comment),
    }

    count = backend.courses_assignments_scores_resolve_and_upload(course_query, assignment_query, scores)

    print(f"Uploaded {count} Scores")

    return 0

def main() -> int:
    """ Get a parser, parse the args, and call run. """
    return run_cli(_get_parser().parse_args())

def _get_parser() -> argparse.ArgumentParser:
    """ Get the parser. """

    parser = lms.cli.parser.get_parser(__doc__.strip(),
            include_course = True,
            include_assignment = True,
            include_user = True,
    )

    parser.add_argument('score', metavar = 'SCORE',
        action = 'store', type = float,
        help = 'Score to upload.')

    parser.add_argument('comment', metavar = 'COMMENT',
        action = 'store', type = str, nargs = '?', default = None,
        help = 'Optional comment.')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import lms.backend.testing
import lms.model.testdata.courses

def test_courses_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of listing courses. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {},
            [
                lms.model.testdata.courses.COURSES['Course 101'],
                lms.model.testdata.courses.COURSES['Course Using Different Languages'],
                lms.model.testdata.courses.COURSES['Extra Course'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_list, test_cases)



import enum
import typing

import lms.model.base
import lms.model.query

class UserQuery(lms.model.query.BaseQuery):
    """
    A class for the different ways one can attempt to reference an LMS user.
    In general, a user can be queried by:
     - LMS User ID (`id`)
     - Email (`email`)
     - Full Name (`name`)
     - f"{email} ({id})"
     - f"{name} ({id})"
    """

    _include_email = True

class ResolvedUserQuery(lms.model.query.ResolvedBaseQuery, UserQuery):
    """
    A UserQuery that has been resolved (verified) from a real user instance.
    """

    _include_email = True

    def __init__(self,
            user: 'ServerUser',
            **kwargs: typing.Any) -> None:
        super().__init__(id = user.id, name = user.name, email = user.email, **kwargs)

class CourseRole(enum.Enum):
    """
    Different roles a user can have in a course.
    LMSs represent this information very differently, so this is only a general collection of roles.
    """

    OTHER = 'other'
    STUDENT = 'student'
    GRADER = 'grader'
    ADMIN = 'admin'
    OWNER = 'owner'

    def __str__(self) -> str:
        return str(self.value)

class ServerUser(lms.model.base.BaseType):
    """
    A user associated with an LMS server.
    """

    CORE_FIELDS = ['id', 'name', 'email']
    """ The common fields shared across backends for this type. """

    def __init__(self,
            id: typing.Union[str, int, None] = None,
            email: typing.Union[str, None] = None,
            name: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        if (id is None):
            raise ValueError("User must have an id.")

        self.id: str = str(id)
        """ The LMS's identifier for this user. """

        self.name: typing.Union[str, None] = name
        """ The display name of this user. """

        self.email: typing.Union[str, None] = email
        """ The email address of this user. """

    def to_query(self) -> ResolvedUserQuery:
        """ Get a query representation of this user. """

        return ResolvedUserQuery(self)

class CourseUser(ServerUser):
    """
    A user associated with a course, e.g., an instructor or student.
    """

    CORE_FIELDS = ServerUser.CORE_FIELDS + ['role']
    """ The common fields shared across backends for this type. """

    def __init__(self,
            role: typing.Union[CourseRole, None] = None,
            raw_role: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        self.role: typing.Union[CourseRole, None] = role
        """ The role of this user within this course (e.g., owner, student). """

        self.raw_role: typing.Union[str, None] = raw_role
        """
        The raw role string from the LMS.
        This may not translate nicely into one of our known roles.
        """

    def is_student(self) -> bool:
        """
        Check if this course user is a student (and therefore be included in graded components like gradebooks).
        Backends should implement this method.
        """

        return (self.role == CourseRole.STUDENT)



import typing

import lms.model.backend
import lms.model.assignments
import lms.model.courses
import lms.model.groupsets
import lms.model.users
import lms.util.parse

def check_required_assignment(
        backend: lms.model.backend.APIBackend,
        config: typing.Dict[str, typing.Any],
        ) -> typing.Union[lms.model.assignments.AssignmentQuery, None]:
    """
    Fetch and ensure that an assignment is provided in the config.
    If no assignment is provided, print a message and return None.
    """

    assignment = lms.util.parse.optional_string(config.get('assignment', None))
    if (assignment is None):
        print('ERROR: No assignment has been provided.')
        return None

    query = backend.parse_assignment_query(assignment)
    if (query is None):
        print('ERROR: Assignment query is malformed.')
        return None

    return query

def check_required_course(
        backend: lms.model.backend.APIBackend,
        config: typing.Dict[str, typing.Any],
        ) -> typing.Union[lms.model.courses.CourseQuery, None]:
    """
    Fetch and ensure that a course is provided in the config.
    If no course is provided, print a message and return None.
    """

    course = lms.util.parse.optional_string(config.get('course', None))
    if (course is None):
        print('ERROR: No course has been provided.')
        return None

    query = backend.parse_course_query(course)
    if (query is None):
        print('ERROR: Course query is malformed.')
        return None

    return query

def check_required_groupset(
        backend: lms.model.backend.APIBackend,
        config: typing.Dict[str, typing.Any],
        ) -> typing.Union[lms.model.groupsets.GroupSetQuery, None]:
    """
    Fetch and ensure that a group set is provided in the config.
    If no group set is provided, print a message and return None.
    """

    groupset = lms.util.parse.optional_string(config.get('groupset', None))
    if (groupset is None):
        print('ERROR: No group set has been provided.')
        return None

    query = backend.parse_groupset_query(groupset)
    if (query is None):
        print('ERROR: Group set query is malformed.')
        return None

    return query

def check_required_group(
        backend: lms.model.backend.APIBackend,
        config: typing.Dict[str, typing.Any],
        ) -> typing.Union[lms.model.groups.GroupQuery, None]:
    """
    Fetch and ensure that a group is provided in the config.
    If no group is provided, print a message and return None.
    """

    group = lms.util.parse.optional_string(config.get('group', None))
    if (group is None):
        print('ERROR: No group has been provided.')
        return None

    query = backend.parse_group_query(group)
    if (query is None):
        print('ERROR: Group query is malformed.')
        return None

    return query

def check_required_user(
        backend: lms.model.backend.APIBackend,
        config: typing.Dict[str, typing.Any],
        ) -> typing.Union[lms.model.users.UserQuery, None]:
    """
    Fetch and ensure that a user is provided in the config.
    If no user is provided, print a message and return None.
    """

    user = lms.util.parse.optional_string(config.get('user', None))
    if (user is None):
        print('ERROR: No user has been provided.')
        return None

    query = backend.parse_user_query(user)
    if (query is None):
        print('ERROR: User query is malformed.')
        return None

    return query
